package sedplugin.characterization;

import java.util.ArrayList;

/**
 * Lists all errors about a {@link Characterization}.
 * 
 * @author Gr&eacute;gory Mantelet (CDS)
 * @version 04/2012
 */
public class CharacErrors {

	public String wlValueError;
	public String wlUnitError;
	public String fluxUnitError;
	public String dateError;

	public boolean wlCalibrated;
	public String wlCalibrationError;

	public boolean fluxCalibrated;
	public String fluxCalibrationError;

	public ArrayList<String> otherErrors;


	public CharacErrors(){ reset(); }

	public CharacErrors(CharacErrors errors){
		if (errors != null){
			wlValueError = (errors.wlValueError==null)?null:new String(errors.wlValueError);
			wlUnitError = (errors.wlUnitError==null)?null:new String(errors.wlUnitError);
			fluxUnitError = (errors.fluxUnitError==null)?null:new String(errors.fluxUnitError);
			dateError = (errors.dateError==null)?null:new String(errors.dateError);

			wlCalibrated = errors.wlCalibrated;
			wlCalibrationError = (errors.wlCalibrationError==null)?null:new String(errors.wlCalibrationError);

			fluxCalibrated = errors.fluxCalibrated;
			fluxCalibrationError = (errors.fluxCalibrationError==null)?null:new String(errors.fluxCalibrationError);

			otherErrors = (errors.otherErrors==null)?new ArrayList<String>():new ArrayList<String>(errors.otherErrors);
		}else
			reset();
	}

	/**
	 * Counts all errors listed in this instance.
	 * 
	 * @return	Number of all Characterization errors.
	 */
	public int countErrors(){
		int nbErrors = otherErrors.size();
		nbErrors += ((wlValueError==null)?0:1)+((wlUnitError==null)?0:1)+((fluxUnitError==null)?0:1)+((dateError==null)?0:1);
		nbErrors += ((wlCalibrationError!=null || !wlCalibrated)?1:0)+((fluxCalibrationError!=null || !fluxCalibrated)?1:0);
		return nbErrors;
	}

	/**
	 * Resets this errors list.
	 */
	public void reset(){
		wlValueError = null;
		wlUnitError = null;
		fluxUnitError = null;
		dateError = null;
		wlCalibrated = true;
		wlCalibrationError = null;
		fluxCalibrated = true;
		fluxCalibrationError = null;
		otherErrors = new ArrayList<String>();
	}

	/**
	 * Tells whether there is no Characterization error, or not.
	 * 
	 * @return	<i>true</i> if there is no error, <i>false</i> otherwise.
	 */
	public boolean isCharacCorrect(){
		return isSpectralCalibrated() && isFluxCalibrated()
		&& wlValueError == null && wlUnitError == null
		&& fluxUnitError == null
		&& dateError == null
		&& otherErrors.isEmpty();
	}

	/**
	 * Tells whether the spectral axis is calibrated or not.
	 * 
	 * @return	<i>true</i> if the spectral axis is calibrated, <i>false</i> otherwise.
	 */
	public boolean isSpectralCalibrated(){
		return wlCalibrated && wlCalibrationError == null;
	}

	/**
	 * Tells whether the flux axis is calibrated or not.
	 * 
	 * @return	<i>true</i> if the flux axis is calibrated, <i>false</i> otherwise.
	 */
	public boolean isFluxCalibrated(){
		return fluxCalibrated && fluxCalibrationError == null;
	}

	/**
	 * Tells whether the Characterization spectral and flux axis are calibrated or not.
	 * 
	 * @return	<i>true</i> if the flux and spectral axis are calibrated, <i>false</i> otherwise.
	 * 
	 * @see #isSpectralCalibrated()
	 * @see #isFluxCalibrated()
	 */
	public boolean isCharacCalibrated(){
		return isSpectralCalibrated() && isFluxCalibrated();
	}

	/**
	 * Formats the list of other Characterization errors (in text or in HTML).
	 * 
	 * @param html	<i>true</i> to format the list of other errors in HTML, <i>false</i> to stay in simple text.
	 * 
	 * @return		The formatted list of the other Characterization errors.
	 */
	public String getGeneralErrors(boolean html){
		String error = html?"<html><ul>":"";

		for(int i=0; i<otherErrors.size(); i++)
			error += (html?"<li>":((i==0)?"- ":"\n- "))+otherErrors.get(i)+(html?"</li>":"");

		if (html)
			error += "</ul></html>";

		return error;
	}

	/**
	 * Formats all the Characterization errors listed in this instance (in text or in HTML).
	 * 
	 * @param html	<i>true</i> to format the all the errors in HTML, <i>false</i> to stay in simple text.
	 * 
	 * @return		The formatted list of all the Characterization errors, or <i>null</i> if there is no error.
	 * 
	 * @see #isCharacCorrect()
	 * @see #isSpectralCalibrated()
	 * @see #isFluxCalibrated()
	 */
	public String getAllErrors(boolean html){
		if (isCharacCorrect())
			return null;

		String allErrors = html?"<html><ul>":"";

		if (!otherErrors.isEmpty()){
			allErrors += (html?"<li>":"")+"General: "+(html?"<ul>":"");
			for(int i=0; i<otherErrors.size(); i++)
				allErrors += (html?"<li>":"\n\t- ")+otherErrors.get(i)+(html?"</li>":"");
			allErrors += html?("</ul></li>"):"\n";
		}

		if (!isSpectralCalibrated() || wlValueError != null || wlUnitError != null){
			allErrors += (html?"<li>":"")+"Wavelength: "+(html?"<ul>":"");
			if (wlCalibrationError != null)
				allErrors += (html?"<li>":"\n\t- ")+"The spectral axis is not calibrated because: "+wlCalibrationError+(html?"</li>":"");
			else if (!wlCalibrated)
				allErrors += (html?"<li>":"\n\t- ")+"The spectral axis is not calibrated !"+(html?"</li>":"");
			if (wlValueError != null)
				allErrors += (html?"<li>":"\n\t- ")+wlValueError+(html?"</li>":"");
			if (wlUnitError != null)
				allErrors += (html?"<li>":"\n\t- ")+wlUnitError+(html?"</li>":"");
			allErrors += (html?"</ul></li>":"\n");
		}

		if (!isFluxCalibrated() || fluxUnitError != null){
			allErrors += (html?"<li>":"")+"Flux: "+(html?"<ul>":"");
			if (fluxCalibrationError != null)
				allErrors += (html?"<li>":"\n\t- ")+"The flux axis is not calibrated because: "+fluxCalibrationError+(html?"</li>":"");
			else if (!fluxCalibrated)
				allErrors += (html?"<li>":"\n\t- ")+"The flux axis is not calibrated !"+(html?"</li>":"");
			if (fluxUnitError != null)
				allErrors += (html?"<li>":"\n\t- ")+fluxUnitError+(html?"</li>":"");
			allErrors += (html?"</ul></li>":"\n");
		}

		if (html)
			allErrors += "</ul></html>";
		else if (allErrors.endsWith("\n"))
			allErrors = allErrors.substring(0, allErrors.length()-1);

		return allErrors;
	}

	/**
	 * Gets a message describing only the calibration errors (only for the spectral and flux axis).
	 * 
	 * @return	A message about the calibration errors or <i>null</i> if there is no calibration error.
	 * 
	 * @see #isCharacCalibrated()
	 */
	public String getCalibrationError() {
		if (isCharacCalibrated())
			return null;
		else{
			String error = "The axis ";
			if (!isSpectralCalibrated() && !isFluxCalibrated())
				error += "\"Spectral\""+((wlCalibrationError!=null)?(" (because: "+wlCalibrationError+")"):"")+" and \"Flux\""+((fluxCalibrationError!=null)?(" (because: "+fluxCalibrationError+")"):"")+" are not ";
			else if (!isSpectralCalibrated())
				error += "\"Spectral\""+((wlCalibrationError!=null)?(" (because: "+wlCalibrationError+")"):"")+" is not ";
			else
				error += "\"Flux\""+((fluxCalibrationError!=null)?(" (because: "+fluxCalibrationError+")"):"")+" is not ";
			return error+"calibrated !";
		}
	}
}
