package sedplugin.characterization;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import sedplugin.errors.CharacterizationException;
import sedplugin.errors.ErrorsLog;
import sedplugin.errors.SEDException;


/**
 * Reads, stores and save a mapping between the Characterization of an image and its FITS header.
 * 
 * @author Gr&eacute;gory Mantelet (CDS)
 * @version 04/2012
 */
public class FitsMapping {

	/* ********* */
	/* CONSTANTS */
	/* ********* */
	public final static String CALIBRATED = "CALIBRATED";
	public final static String NOT_CALIBRATED = "UNCALIBRATED";

	public final static String SPECTRAL_AXIS_UTYPE = "SpectralAxis";
	public final static String SPECTRAL_CALIBRATION_UTYPE = SPECTRAL_AXIS_UTYPE+".calibrationStatus";
	public final static String WAVELENGTH_VALUE_UTYPE = SPECTRAL_AXIS_UTYPE+".Coverage.Location.Coord.Spectral.Value";
	public final static String WAVELENGTH_UNIT_UTYPE = SPECTRAL_AXIS_UTYPE+".Unit";

	public final static String FLUX_AXIS_UTYPE = "FluxAxis";
	public final static String FLUX_CALIBRATION_UTYPE = FLUX_AXIS_UTYPE+".calibrationStatus";
	public final static String FLUX_UNIT_UTYPE = FLUX_AXIS_UTYPE+".Unit";

	public final static String TIME_AXIS_UTYPE = "TimeAxis";
	public final static String DATE_UTYPE = TIME_AXIS_UTYPE+".Coverage.Location.Coord.Time.TimeInstant.MJDTime";

	public final static String VALUE_KEY_SEPARATOR = " # ";

	public final static String VALUE_PREFIX = "~";
	public final static String FITS_KEYWORD_PREFIX = "$";

	/* ********* */
	/* VARIABLES */
	/* ********* */
	protected String spectralCalibration = CALIBRATED;
	protected String wavelength = "";
	protected String wavelengthUnit = "";
	protected String fluxCalibration = CALIBRATED;
	protected String fluxUnit = "";
	protected String date = "";

	protected File mappingFile;
	protected boolean saved;

	/* ************ */
	/* CONSTRUCTORS */
	/* ************ */
	public FitsMapping(){
		mappingFile = null;
		saved = false;
	}

	public FitsMapping(File mappingFile) throws CharacterizationException {
		setMappingFile(mappingFile);
	}

	public FitsMapping(Characterization charac) {
		this();

		if (charac != null){
			spectralCalibration = charac.getSpectralCalibration();
			wavelength = charac.getWavelength()+"";
			wavelengthUnit = charac.getWavelengthUnit();
			fluxCalibration = charac.getFluxCalibration();
			fluxUnit = charac.getFluxUnit();
			date = charac.getDate();
		}
	}

	public FitsMapping(FitsMapping mappingToCopy) {
		this();
		if (mappingToCopy != null){
			spectralCalibration = mappingToCopy.getSpectralCalibration();
			wavelength = mappingToCopy.getWavelength();
			wavelengthUnit = mappingToCopy.getWavelengthUnit();
			fluxCalibration = mappingToCopy.getFluxCalibration();
			fluxUnit = mappingToCopy.getFluxUnit();
			date = mappingToCopy.getDate();

			if (mappingToCopy.getMappingFile() != null)
				mappingFile = new File(mappingToCopy.getMappingFile().getAbsolutePath());
			else
				mappingFile = null;
			saved = true;
		}
	}

	/* ******* */
	/* METHODS */
	/* ******* */
	/**
	 * Reads the associated FITS mapping file (if any) and stores all the data useful to the SEDPlugin.
	 * 
	 * @throws CharacterizationException	If the associated file does not exist, is not file or if there is an error while reading it.
	 */
	public void reload() throws CharacterizationException {
		if (mappingFile != null){
			if (!mappingFile.exists())
				throw new CharacterizationException(null, mappingFile.getAbsolutePath(), false, CharacterizationException.READING_ERROR, "Impossible to load the manual Characterization because the file doesn't exist !");
			if (!mappingFile.isFile())
				throw new CharacterizationException(null, mappingFile.getAbsolutePath(), false, CharacterizationException.READING_ERROR, "Impossible to load the manual Characterization because the given file is not a regular file !");

			// Reset the calibration flags:
			spectralCalibration = CALIBRATED;
			fluxCalibration = CALIBRATED;

			// Open an input stream:
			int nbReadLines = 0;
			try{
				BufferedReader reader = new BufferedReader(new FileReader(mappingFile));
				String line = "", axis = null;
				while((line = reader.readLine()) != null){
					nbReadLines++;
					if (line.trim().length() == 0){
						axis = null;
						continue;
					}
					if (line.indexOf(VALUE_KEY_SEPARATOR) < 0)
						throw generateSyntaxError(nbReadLines, mappingFile.getAbsolutePath(), "The key/value separator char (\""+VALUE_KEY_SEPARATOR+"\") is missing.");
					String[] fields = line.split(VALUE_KEY_SEPARATOR);
					fields[0] = fields[0].trim();
					fields[1] = fields[1].trim();
					if (fields[1].indexOf(".") > 0){
						if (axis == null)
							throw generateSyntaxError(nbReadLines, mappingFile.getAbsolutePath(), "Axis declaration missing.");
						else if (!fields[1].toLowerCase().startsWith(axis.toLowerCase()))
							throw generateSyntaxError(nbReadLines, mappingFile.getAbsolutePath(), "This section is dedicated to the axis \""+axis+"\" so all declared UType must be prefixed by this axis (for instance: "+axis+".Unit).");
						else{
							if (fields[1].equalsIgnoreCase(SPECTRAL_CALIBRATION_UTYPE))
								spectralCalibration = fields[0];
							else if (fields[1].equalsIgnoreCase(WAVELENGTH_VALUE_UTYPE))
								wavelength = fields[0];
							else if (fields[1].equalsIgnoreCase(WAVELENGTH_UNIT_UTYPE))
								wavelengthUnit = fields[0];
							else if (fields[1].equalsIgnoreCase(FLUX_CALIBRATION_UTYPE))
								fluxCalibration = fields[0];
							else if (fields[1].equalsIgnoreCase(FLUX_UNIT_UTYPE))
								fluxUnit = fields[0];
							else if (fields[1].equalsIgnoreCase(DATE_UTYPE))
								date = fields[0];
						}
					}else
						axis = fields[1];
				}
				reader.close();
			}catch(IOException ioe){
				throw new CharacterizationException(null, mappingFile.getAbsolutePath(), false, CharacterizationException.PARSING_ERROR, nbReadLines, ioe);
			}
		}
		saved = (mappingFile != null);
	}

	/**
	 * Saves the FITS mapping represented by this object in the associated file.
	 * If no file is associated with this mapping, this function returns <i>false</i> immediately.
	 * 
	 * @return	<i>true</i> if this mapping has been successfully saved, <i>false</i> otherwise.
	 * 
	 * @throws CharacterizationException	If there is an error while writing this FITS mapping.
	 */
	public boolean save() throws CharacterizationException {
		if (mappingFile == null)
			return false;

		// Create the HashMap of value to update:
		HashMap<String, HashMap<String, String>> mapAxis = generateMappingMap();
		int numMaxAxis = -1;

		String mapping = null;
		int nbReadLines = 0;
		int nbWrittenLines = 0;

		try{
			// Extract the existent file and update lines which contain a UType to update:
			if (mappingFile.exists()){
				if (!mappingFile.isFile())
					throw new CharacterizationException(null, mappingFile.getAbsolutePath(), false, CharacterizationException.WRITTING_ERROR, "Impossible to save the manual Characterization because because the given file is not a regular file !");
				else{
					// Build the reader:
					BufferedReader reader = new BufferedReader(new FileReader(mappingFile));
					String line = "", axis = null;
					HashMap<String, String> currentAxis = null;		// UTypes to update in the current axis
					// For each line....
					while((line = reader.readLine()) != null){
						nbReadLines++;
						// an empty line means the end of an axis, so all remaining UTypes must be added:
						if (line.trim().length()==0){
							// add all remaining UTypes for this axis:
							if (currentAxis != null){
								Iterator<Map.Entry<String, String>> itField = currentAxis.entrySet().iterator();
								while(itField.hasNext()){
									Map.Entry<String, String> field = itField.next();
									mapping = ((mapping==null)?"":(mapping+"\n"))+field.getValue()+VALUE_KEY_SEPARATOR+field.getKey();
								}
							}
							mapAxis.remove(axis);
							// forget the axis:
							axis = null;
							currentAxis = null;
							if (mapping != null)
								mapping += "\n";

						}else{
							// no separator between value and UType => ERROR:
							if (line.indexOf(VALUE_KEY_SEPARATOR) < 0)
								throw generateSyntaxError(nbReadLines, mappingFile.getAbsolutePath(), "The key/value separator char (\""+VALUE_KEY_SEPARATOR+"\") is missing.");
							// Split the UType and its value:
							String[] fields = line.split(VALUE_KEY_SEPARATOR);
							fields[0] = fields[0].trim();
							fields[1] = fields[1].trim();
							// CASE: UType of an axis:
							if (fields[1].indexOf(".") <= 0){
								// update axis:
								if (fields[1].equalsIgnoreCase(SPECTRAL_AXIS_UTYPE))
									axis = SPECTRAL_AXIS_UTYPE;
								else if (fields[1].equalsIgnoreCase(FLUX_AXIS_UTYPE))
									axis = FLUX_AXIS_UTYPE;
								else if (fields[1].equalsIgnoreCase(TIME_AXIS_UTYPE))
									axis = TIME_AXIS_UTYPE;
								else
									axis = fields[1];
								// update currentAxis:
								currentAxis = mapAxis.get(axis);
								// update the maximum of the axis numbers:
								try{
									int numAxis = Integer.parseInt(extractValue(fields[0]));
									numMaxAxis = Math.max(numMaxAxis, numAxis);
								}catch(NumberFormatException nfe){ ErrorsLog.getCurrentInstance().addException(new SEDException(nfe, "While saving a FITS mapping")); }
								// CASE: other UTypes:
							}else{
								// axis not known = axis not declared => ERROR:
								if (axis == null)
									throw generateSyntaxError(nbReadLines, mappingFile.getAbsolutePath(), "Axis declaration missing.");
								// axis of the UType <> known axis => ERROR:
								else if (!fields[1].toLowerCase().startsWith(axis.toLowerCase()))
									throw generateSyntaxError(nbReadLines, mappingFile.getAbsolutePath(), "This section is dedicated to the axis \""+axis+"\" so all declared UType must be prefixed by this axis (for instance: "+axis+".Unit).");
								else{
									String newValue = fields[0];
									if (fields[1].equalsIgnoreCase(SPECTRAL_CALIBRATION_UTYPE) && currentAxis.containsKey(SPECTRAL_CALIBRATION_UTYPE)){
										newValue = spectralCalibration;
										currentAxis.remove(SPECTRAL_CALIBRATION_UTYPE);
									}else if (fields[1].equalsIgnoreCase(WAVELENGTH_VALUE_UTYPE) && currentAxis.containsKey(WAVELENGTH_VALUE_UTYPE)){
										newValue = wavelength;
										currentAxis.remove(WAVELENGTH_VALUE_UTYPE);
									}else if (fields[1].equalsIgnoreCase(WAVELENGTH_UNIT_UTYPE) && currentAxis.containsKey(WAVELENGTH_UNIT_UTYPE)){
										newValue = wavelengthUnit;
										currentAxis.remove(WAVELENGTH_UNIT_UTYPE);
									}else if (fields[1].equalsIgnoreCase(FLUX_CALIBRATION_UTYPE) && currentAxis.containsKey(FLUX_CALIBRATION_UTYPE)){
										newValue = fluxCalibration;
										currentAxis.remove(FLUX_CALIBRATION_UTYPE);
									}else if (fields[1].equalsIgnoreCase(FLUX_UNIT_UTYPE) && currentAxis.containsKey(FLUX_UNIT_UTYPE)){
										newValue = fluxUnit;
										currentAxis.remove(FLUX_UNIT_UTYPE);
									}else if (fields[1].equalsIgnoreCase(DATE_UTYPE) && currentAxis.containsKey(DATE_UTYPE)){
										newValue = date;
										currentAxis.remove(DATE_UTYPE);
									}
									line = newValue+VALUE_KEY_SEPARATOR+fields[1];
								}
							}
							mapping = ((mapping==null)?"":(mapping+"\n"))+line;
						}
					}
					reader.close();
				}

			}

			// If some axis have not been updated, add them at the end of the file:
			if (!mapAxis.isEmpty()){
				Iterator<Map.Entry<String, HashMap<String, String>>> itAxis = mapAxis.entrySet().iterator();
				while(itAxis.hasNext()){
					Map.Entry<String, HashMap<String, String>> axisDef = itAxis.next();
					if (axisDef.getValue().size() > 0){
						mapping = ((mapping==null)?"":(mapping+"\n\n"))+VALUE_PREFIX+(++numMaxAxis)+VALUE_KEY_SEPARATOR+axisDef.getKey();
						Iterator<Map.Entry<String, String>> itDef = axisDef.getValue().entrySet().iterator();
						while(itDef.hasNext()){
							Map.Entry<String, String> def = itDef.next();
							mapping += "\n"+def.getValue()+VALUE_KEY_SEPARATOR+def.getKey();
						}
					}
				}
			}

			// Save the mapping file:
			BufferedWriter writer = new BufferedWriter(new FileWriter(mappingFile));
			writer.write(mapping);
			writer.close();
		} catch(IOException ioe) {
			throw new CharacterizationException(null, mappingFile.getAbsolutePath(), false, (nbWrittenLines==0)?nbReadLines:nbWrittenLines, (nbWrittenLines==0)?CharacterizationException.READING_ERROR:CharacterizationException.WRITTING_ERROR, ioe);
		}

		return saved = true;
	}

	/**
	 * Saves this FITS mapping in the given file.
	 * 
	 * @param file		File in which this FITS mapping must be saved.
	 * 
	 * @return			<i>true</i> if this FITS mapping has been successfully saved, <i>false</i> otherwise.
	 * 
	 * @throws CharacterizationException	If there is an error while writing this FITS mapping.
	 * 
	 * @see #save()
	 */
	public boolean save(File file) throws CharacterizationException {
		if (file != null){
			mappingFile = file;
			saved = false;
			return save();
		}else
			return false;
	}

	/**
	 * Tells whether this FITS mapping is saved or not.
	 * 
	 * @return	<i>true</i> if this FITS mapping is saved, <i>false</i> otherwise.
	 */
	public final boolean isSaved(){
		return saved;
	}

	/**
	 * Generates a map in which all interesting FITS keywords will be saved while reading a FITS mapping file.
	 * 
	 * @return	A map in which keys are axis names (spectral, flux and time),
	 * 			and values are the axis fields (calibration, value, unit).
	 */
	protected HashMap<String, HashMap<String, String>> generateMappingMap(){
		HashMap<String, HashMap<String, String>> mapAxis = new HashMap<String, HashMap<String, String>>();

		HashMap<String, String> mapSpectralAxis = new HashMap<String, String>();
		mapSpectralAxis.put(SPECTRAL_CALIBRATION_UTYPE, spectralCalibration);
		mapSpectralAxis.put(WAVELENGTH_VALUE_UTYPE, wavelength);
		mapSpectralAxis.put(WAVELENGTH_UNIT_UTYPE, wavelengthUnit);
		mapAxis.put(SPECTRAL_AXIS_UTYPE, mapSpectralAxis);

		HashMap<String, String> mapFluxAxis = new HashMap<String, String>();
		mapFluxAxis.put(FLUX_CALIBRATION_UTYPE, fluxCalibration);
		mapFluxAxis.put(FLUX_UNIT_UTYPE, fluxUnit);
		mapAxis.put(FLUX_AXIS_UTYPE, mapFluxAxis);

		HashMap<String, String> mapTimeAxis = new HashMap<String, String>();
		mapTimeAxis.put(DATE_UTYPE, date);
		mapAxis.put(TIME_AXIS_UTYPE, mapTimeAxis);

		return mapAxis;
	}

	/* ************ */
	/* TOOL METHODS */
	/* ************ */
	/**
	 * Generates a Characterization exception describing a syntax error in the specified FITS mapping file,
	 * at the given line and with the given explanation.
	 * 
	 * @param line		Line of the file at which the error occurs.
	 * @param file		File in which the error occurs.
	 * @param message	Explanation of the error.
	 * 
	 * @return			The corresponding {@link CharacterizationException}.
	 */
	protected static CharacterizationException generateSyntaxError(int line, String file, String message){
		return new CharacterizationException(null, file, false, line, message);
	}

	/**
	 * Tells whether the given mapping value is a FITS keyword.
	 * A FITS keyword starts with {@link #FITS_KEYWORD_PREFIX}.
	 * 
	 * @param value	A mapping value.
	 * 
	 * @return		<i>true</i> if the given mapping value is a FITS keyword, <i>false</i> otherwise.
	 */
	public static boolean isFitsKeyword(String value){
		return value.startsWith(FITS_KEYWORD_PREFIX);
	}

	/**
	 * Gets the name of the FITS keyword specified in the given mapping value.
	 * 
	 * @param value	A mapping value.
	 * 
	 * @return		The found FITS keyword or <i>null</i> if the given value is a constant and not a FITS keyword.
	 * 
	 * @see #isFitsKeyword(String)
	 */
	public static String extractFitsKeywordName(String value){
		return isFitsKeyword(value)?value.substring(value.indexOf(FITS_KEYWORD_PREFIX)+1):null;
	}

	/**
	 * Builds the mapping value to write in a FITS mapping file for a FITS keyword.
	 * For instance, if the FITS keyword is "foo", {@link #FITS_KEYWORD_PREFIX}+"foo" will be returned.
	 * 
	 * @param keywordName	Name of a FITS keyword.
	 * 
	 * @return				Encoded FITS keyword.
	 */
	public static String getFitsKeyword(String keywordName){
		return isFitsKeyword(keywordName)?keywordName:(FITS_KEYWORD_PREFIX+keywordName);
	}

	/**
	 * Extracts the value from the given mapping value.
	 * A constant value starts with {@link #VALUE_PREFIX}.
	 * 
	 * @param value	A mapping value.
	 * 
	 * @return		The constant value.
	 */
	public static String extractValue(String value){
		return value.substring(value.indexOf(VALUE_PREFIX)+1);
	}

	/**
	 * Builds the mapping value to write in a FITS mapping file for a constant value.
	 * For instance, if the constant is "foo", {@link #VALUE_PREFIX}+"foo" will be returned.
	 * 
	 * @param value	Constant to encode.
	 * 
	 * @return		Encoded constant value.
	 */
	public static String getValue(String value){
		return value.startsWith(VALUE_PREFIX)?value:(VALUE_PREFIX+value);
	}

	/* ******************* */
	/* GETTERS AND SETTERS */
	/* ******************* */
	/**
	 * @return The mappingUrl.
	 */
	public final File getMappingFile() {
		return mappingFile;
	}
	/**
	 * Load the given mapping file and update this FitsMapping object.
	 * @param mappingUrl The mappingUrl to set.
	 */
	public final void setMappingFile(File mappingFile) throws CharacterizationException {
		this.mappingFile = mappingFile;
		reload();
	}
	/**
	 * @return
	 */
	public final String getSpectralCalibration(){
		return spectralCalibration;
	}
	/**
	 * @param calibrated
	 */
	public final void setSpectralCalibration(String calibration){
		spectralCalibration = (calibration == null)?"":calibration.trim();
		saved = false;
	}
	/**
	 * @return The wavelength.
	 */
	public final String getWavelength() {
		return extractValue(wavelength);
	}
	/**
	 * @param wavelength The wavelength to set.
	 */
	public final void setWavelength(String wavelength) {
		this.wavelength = (wavelength==null)?"":wavelength.trim();
		saved = false;
	}
	/**
	 * @return The wavelengthUnit.
	 */
	public final String getWavelengthUnit() {
		return extractValue(wavelengthUnit);
	}
	/**
	 * @param wavelengthUnit The wavelengthUnit to set.
	 */
	public final void setWavelengthUnit(String wavelengthUnit) {
		this.wavelengthUnit = (wavelengthUnit==null)?"":wavelengthUnit.trim();
		saved = false;
	}
	/**
	 * @return
	 */
	public final String getFluxCalibration(){
		return fluxCalibration;
	}
	/**
	 * @param calibrated
	 */
	public final void setFluxCalibration(String calibration){
		fluxCalibration = (calibration == null)?"":calibration.trim();
		saved = false;
	}
	/**
	 * @return The fluxUnit.
	 */
	public final String getFluxUnit() {
		return extractValue(fluxUnit);
	}
	/**
	 * @param fluxUnit The fluxUnit to set.
	 */
	public final void setFluxUnit(String fluxUnit) {
		this.fluxUnit = (fluxUnit==null)?"":fluxUnit.trim();
		saved = false;
	}
	/**
	 * @return The date.
	 */
	public final String getDate() {
		return extractValue(date);
	}
	/**
	 * @param date The date to set.
	 */
	public final void setDate(String date) {
		this.date = (date==null)?"":date.trim();
		saved = false;
	}

	@Override
	public String toString(){
		return "*** Characterization ["+mappingFile+"] ["+(isSaved()?"SAVED":"NOT SAVED")+"] ***\n\t- Wavelength: "+getWavelength()+"\n\t- Wavelength unit: "+getWavelengthUnit()+"\n\t- Flux unit: "+getFluxUnit()+"\n\t- Date: "+getDate()+"\n";
	}

	/* ************************ */
	/* MAIN METHODS (SELF TEST) */
	/* ************************ */
	public final static void main(String[] args) throws Exception {
		File mappingFile = new File("mappingTest.map");
		FitsMapping mapping;
		if (mappingFile.exists())
			mapping = new FitsMapping(mappingFile);
		else
			mapping = new FitsMapping();
		System.out.println(mapping);
		mapping.setWavelength(getValue("60"));
		mapping.setWavelengthUnit(getValue("mu"));
		System.out.println("\n"+mapping);
		System.out.println("mappingTestSave.map: "+(mapping.save()?"SAVED":"NOT SAVED"));
		System.out.println("mappingTestSave2.map :"+(mapping.save(new File("mappingTestSave2.map"))?"SAVED":"NOT SAVED"));
	}

}
