package sedplugin.characterization;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;

import java.net.URL;

import java.text.ParseException;
import java.text.SimpleDateFormat;

import java.util.Date;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Vector;

import sedplugin.errors.ErrorsLog;
import sedplugin.errors.SEDException;

import cds.astro.Astrotime;

/**
 * Content of a FITS header.
 * 
 * @author Gr&eacute;gory Mantelet (CDS)
 * @version 04/2012
 */
public class HeaderFits {

	/* ********* */
	/* CONSTANTS */
	/* ********* */
	public static final int UNDEFINED = -1;
	public static final int EMPTY_STRING = 0;
	public static final int STRING = 1;
	public static final int LOGICAL = 2;
	public static final int NUMERIC = 3;

	/* ***************** */
	/* CLASS HEADER_ITEM */
	/* ***************** */
	/**
	 * A FITS header item/row.
	 * 
	 * @author Gr&eacute;gory Mantelet (CDS)
	 * @version 04/2012
	 */
	public static class HeaderItem {
		public final String keyword;
		public final String value;
		public final int type;
		public final String comment;

		public HeaderItem(String keyword, String value, String comment){
			this.keyword = keyword;
			value = value.trim();
			this.type = identifyValueType(value.trim());
			if (this.type == STRING)
				this.value = value.substring(1, value.length()-1).trim();
			else
				this.value = value;
			this.comment = comment;
		}

		@Override
		public String toString(){
			return keyword+"{"+getTypeString(type)+"} = "+value+" ["+comment+"]";
		}
	}

	/* ********** */
	/* ATTRIBUTES */
	/* ********** */
	protected LinkedList<HeaderItem> keywords;
	protected HashMap<String, HeaderItem> mapKeywords;

	/* ************ */
	/* CONSTRUCTORS */
	/* ************ */
	/**
	 * Builds an empty FITS header.
	 * It can be filled later thanks to {@link #loadHeader(InputStream)} and {@link #loadHeader(String)}.
	 */
	public HeaderFits(){
		keywords = new LinkedList<HeaderItem>();
		mapKeywords = new HashMap<String, HeaderItem>();
	}

	/**
	 * Loads the header of the specified FITS file.
	 * 
	 * @param imageURL		URL of a FITS file.
	 * 
	 * @throws IOException	If there is an error while reading the specified FITS file.
	 * 
	 * @see #loadHeader(InputStream)
	 */
	public HeaderFits(URL imageURL) throws IOException{
		loadHeader(imageURL.openStream());
	}

	/**
	 * Loads the header of the given FITS file.
	 * 
	 * @param image			FITS file.
	 * 
	 * @throws FileNotFoundException	If the given file does not exist.
	 * @throws IOException				If there is an error while reading the given FITS file.
	 * 
	 * @see #loadHeader(InputStream)
	 */
	public HeaderFits(File image) throws FileNotFoundException, IOException{
		loadHeader(new FileInputStream(image));
	}

	/**
	 * Loads the given FITS header.
	 * 
	 * @param header		Content or part of the content of a FITS header.
	 * 
	 * @throws IOException	If the given content is empty.
	 * 
	 * @see #loadHeader(String)
	 */
	public HeaderFits(String header) throws IOException{
		if (header == null)
			throw new IOException("Empty header !");
		loadHeader(header);
	}

	/* ******* */
	/* METHODS */
	/* ******* */
	/**
	 * Gets the FITS item type: {@link #EMPTY_STRING}="EMPTY", {@link #STRING}="STRING", {@link #LOGICAL}="LOGICAL",
	 * {@link #NUMERIC}="NUMERIC", any other value will be returned as "UNDEFINED".
	 * 
	 * @param type	A FITS item type.
	 * 
	 * @return		The corresponding type string.
	 */
	public static String getTypeString(int type){
		switch(type){
		case EMPTY_STRING:
			return "EMPTY";
		case STRING:
			return "STRING";
		case LOGICAL:
			return "LOGICAL";
		case NUMERIC:
			return "NUMERIC";
		case UNDEFINED:
		default:
			return "UNDEFINED";
		}
	}

	/**
	 * Identifies the FITS item type from its value:
	 * T or F = {@link #LOGICAL}, '...' = {@link #EMPTY_STRING} or {@link #STRING},
	 * other values are tested as numeric ({@link #NUMERIC}), but if the test fails, {@link #UNDEFINED}
	 * is returned.
	 * 
	 * @param value	Value whose the FITS type must be found.
	 * 
	 * @return		The corresponding FITS type.
	 */
	public static int identifyValueType(String value){
		if (value.toUpperCase().equals("T") || value.toUpperCase().equals("F"))
			return LOGICAL;
		else if (value.startsWith("'") && value.endsWith("'")){
			if (value.substring(1, value.length()-1).trim().length()==0)
				return EMPTY_STRING;
			else
				return STRING;
		}else{
			try{
				Double.parseDouble(value);
				return NUMERIC;
			}catch(NumberFormatException nfe){
				return UNDEFINED;
			}
		}
	}

	/**
	 * Gets all loaded FITS keywords.
	 * 
	 * @return	An iterator on all loaded FITS keywords.
	 */
	public final Iterator<HeaderItem> getKeywords(){
		return keywords.iterator();
	}

	/**
	 * Gets all loaded FITS keywords.
	 * 
	 * @return	A vector with all loaded FITS keywords.
	 */
	public Vector<HeaderItem> getAllKeywords(){
		Vector<HeaderItem> items = new Vector<HeaderItem>(keywords.size());
		Iterator<HeaderItem> it = getKeywords();
		while(it.hasNext())
			items.add(it.next());
		return items;
	}

	/**
	 * Gets all loaded FITS keywords which are of the given type.
	 * 
	 * @param type	FITS item type ( {@link #EMPTY_STRING} {@link #STRING}, {@link #NUMERIC}, {@link #LOGICAL}, {@link #UNDEFINED}).
	 * 
	 * @return		All the corresponding FITS keywords.
	 */
	public Vector<HeaderItem> getAllKeywords(int type){
		Vector<HeaderItem> items = new Vector<HeaderItem>();
		Iterator<HeaderItem> it = getKeywords();
		while(it.hasNext()){
			HeaderItem item = (HeaderItem)it.next();
			if (item.type == type)
				items.add(item);
		}
		return items;
	}

	/**
	 * Gets the whole FITS item whose the keyword matches exactly (case sensitive) to the given one.
	 * 
	 * @param keyword	FITS keyword.
	 * 
	 * @return			The corresponding FITS item.
	 */
	public HeaderItem getKeywordItem(String keyword){
		return (HeaderItem)mapKeywords.get(keyword);
	}

	/**
	 * Gets the number of all loaded FITS keywords.
	 * 
	 * @return	The number of all loaded FITS keywords.
	 */
	public final int getNbKeywords(){
		return keywords.size();
	}

	/**
	 * Loads the given FITS header.
	 * 
	 * @param headerFits	FITS header to load.
	 */
	public void loadHeader(String headerFits) {
		keywords = new LinkedList<HeaderItem>();
		mapKeywords = new HashMap<String, HeaderItem>();

		String[] lines = headerFits.split("\n");
		boolean end = false;
		for(int i=0; !end && i<lines.length; i++){
			// Extract the fits keyword:
			String line = lines[i];
			String key = (line.substring(0, 8)).trim();
			if (!(end = key.equals("END"))){
				// Extract the remaining string:
				String value = (line.substring(10)).trim();
				String comment = null;
				// Search the good index of the comment beginning (so that a string value between two ' can contain / characters):
				int indComment = value.indexOf("/");
				int indApost = value.indexOf("'");
				if (indApost > -1 && indApost < indComment){
					indApost = value.substring(1).indexOf("'");
					indComment = value.substring(indApost+1).indexOf("/")+indApost+1;
				}
				// Extract the value and the comment:
				if (indComment >= 0){
					comment = value.substring(indComment+1).trim();
					value = value.substring(0, indComment).trim();
				}
				// Build the header item;
				HeaderItem headerItem = new HeaderItem(key, value, comment);
				// Save it:
				keywords.add(headerItem);
				mapKeywords.put(key, headerItem);
			}
		}
	}

	/**
	 * Loads the FITS header from the given stream on an FITS image.
	 * 
	 * @param is			Input stream on a FITS image.
	 * 
	 * @throws IOException	If there is an error while reading the given stream.
	 */
	public void loadHeader(InputStream is) throws IOException {
		byte[] line = new byte[80];
		String remaining = "";
		boolean end = false;

		keywords = new LinkedList<HeaderItem>();
		mapKeywords = new HashMap<String, HeaderItem>();

		while (!end && is.read(line) >= 0){
			String lineStr = remaining+new String(line);

			int indReturn = lineStr.indexOf("\n");
			if (indReturn >= 0){
				remaining = lineStr.substring(indReturn+1);
				lineStr = lineStr.substring(0, indReturn);
			}else
				remaining = "";

			// Extract the fits keyword:
			String key = (lineStr.substring(0, 8)).trim();
			if (!(end = key.equals("END"))){
				// Extract the remaining string:
				String value = (lineStr.substring(10)).trim();
				String comment = null;
				// Search the good index of the comment beginning (so that a string value between two ' can contain / characters):
				int indComment = value.indexOf("/");
				int indApost = value.indexOf("'");
				if (indApost > -1 && indApost < indComment){
					indApost = value.substring(1).indexOf("'");
					indComment = value.substring(indApost+1).indexOf("/")+indApost+1;
				}
				// Extract the value and the comment:
				if (indComment >= 0){
					comment = value.substring(indComment+1).trim();
					value = value.substring(0, indComment).trim();
				}
				// Build the header item;
				HeaderItem headerItem = new HeaderItem(key, value, comment);
				// Save it:
				keywords.add(headerItem);
				mapKeywords.put(key, headerItem);
			}
		}
	}

	/**
	 * Converts the given date string into a MJD date.
	 * 
	 * @param dateStr	A date (expected format: dd/MM/yy).
	 * 
	 * @return			The corresponding MJDdate.
	 * 
	 * @see Astrotime#JD2MJD(double)
	 */
	protected static double dateToMJD(String dateStr){
		String dateFormat = "dd/MM/yy";
		SimpleDateFormat dateParser = new SimpleDateFormat(dateFormat);
		try {
			Date date = dateParser.parse(dateStr.toLowerCase());
			GregorianCalendar calendar = new GregorianCalendar();
			calendar.setTime(date);
			double mjdDate = Astrotime.JD2MJD(Astrotime.ms2JD(calendar.getTimeInMillis()));
			return mjdDate;
		} catch (ParseException e) {
			ErrorsLog.getCurrentInstance().addException(new SEDException("Incorrect date format", "HeaderFits.dateToMJD", "The date format \""+dateFormat+"\" is incorrect for the given date \""+dateStr+"\" !"));
			return -1;
		}
	}

	/* *********************** */
	/* MAIN METHOD (SELF TEST) */
	/* *********************** */
	public static void main(String[] args) throws IOException {
		String fitsIMAGE = "...";
		HeaderFits reader = new HeaderFits();

		System.out.println("Parsing header FITS....");
		reader.loadHeader(new FileInputStream(new File(fitsIMAGE)));
		System.out.println("Parsing FINISHED !\n\n*** HEADER ***");

		for(int i=0; i<reader.keywords.size(); i++){
			HeaderItem item = (HeaderItem)reader.keywords.get(i);
			System.out.println(item);
		}

		System.out.print("SEARCHING [QPOENAME]...");
		HeaderItem item = reader.getKeywordItem("QPOENAME");
		System.out.println((item==null)?"NOT FOUND":("FOUND {"+item+"}"));
	}

}
