package sedplugin.interfaceGraphique;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.net.URL;
import java.util.Iterator;
import java.util.Vector;

import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.ListCellRenderer;
import javax.swing.WindowConstants;


import sedplugin.errors.SEDException;
import sedplugin.plugin.SEDPlugin;
import sedplugin.sed.SED;
import sedplugin.sed.source.SEDSource;
import sedplugin.sed.source.aladin.Circle;
import sedplugin.sed.source.aladin.Polygon;
import sedplugin.sed.source.operation.Operation;

/**
 * A simple dialog to make difference between two spectrums.<BR />
 * If flux axis is in FLUX quantity then the difference will be done after converting the flux axis in INTENSITY quantity.<BR />
 * If not changed by the user the default name of the resulting spectrum is "On - Off" followed by the number of done subtractions.
 *
 * @author Gregory Mantelet
 * @version 03/2012
 */
public class BackgroundSubtractorDialog extends JDialog {

	private static final long serialVersionUID = 1L;

	protected final SEDPlugin plugin;
	protected Vector<SED> seds = new Vector<SED>();

	protected static int numSubtraction = 0;

	protected JTextField txtResultName;

	protected JComboBox cboLeftOperand;
	protected JComboBox cboRightOperand;

	protected JButton cmdCompute;
	protected JButton cmdCancel;

	/** Icon for a spectrum coming from a position selection. */
	private static ImageIcon icoPos = null;

	/** Icon for a spectrum coming from a circle selection. */
	private static ImageIcon icoCircle = null;

	/** Icon for a spectrum coming from a polygon selection. */
	private static ImageIcon icoPolygon = null;

	/** Icon for a spectrum coming from a spectrums operation. */
	private static ImageIcon icoOperation = null;

	/** Icon for a spectrum coming from a catalog. */
	private static ImageIcon icoCatalog = null;

	/* ************ */
	/* CONSTRUCTORS */
	/* ************ */
	public BackgroundSubtractorDialog(JFrame owner, SEDPlugin plug){
		super(owner);

		// Manage parameters:
		plugin = plug;

		// Build GI:
		build();

		// Display dialog:
		setLocationRelativeTo(owner);
		setVisible(true);
	}

	/* **************** */
	/* BUILDING METHODS */
	/* **************** */
	private void build(){
		buildSEDIcon();

		// RESULT NAME PANEL:
		JLabel lblName = new JLabel("<html><b>Name</b> <i>(needed)</i>: ");

		txtResultName = new JTextField("On - Off ("+numSubtraction+")");
		txtResultName.addKeyListener(new KeyListener(){
			public void keyReleased(KeyEvent e) {
				cmdCompute.setEnabled(txtResultName.getText().trim().length()!=0);
			}

			public void keyTyped(KeyEvent arg0) {;}
			public void keyPressed(KeyEvent arg0) {;}
		});

		JPanel panResultName = new JPanel(new BorderLayout());
		panResultName.add(lblName, BorderLayout.WEST);
		panResultName.add(txtResultName, BorderLayout.CENTER);

		// OPERATION PANEL:
		seds = new Vector<SED>();
		Iterator<SED> it = plugin.getSEDs();
		while(it.hasNext())
			seds.add(it.next());

		cboLeftOperand = new JComboBox(seds);
		cboLeftOperand.setRenderer(new SEDListCellRenderer());
		cboLeftOperand.setEditable(false);
		if (seds.size()> 0) cboLeftOperand.setSelectedIndex(0);

		JLabel lblMinus = new JLabel(" - ");
		lblMinus.setFont(lblMinus.getFont().deriveFont(Font.BOLD));

		cboRightOperand = new JComboBox(seds);
		cboRightOperand.setRenderer(new SEDListCellRenderer());
		cboRightOperand.setEditable(false);
		if (seds.size() > 0) cboRightOperand.setSelectedIndex((seds.size()>1)?1:0);

		JPanel panOperation = new JPanel();
		panOperation.add(cboLeftOperand);
		panOperation.add(lblMinus);
		panOperation.add(cboRightOperand);

		// COMMANDS PANEL:
		cmdCompute = new JButton("Compute");
		cmdCompute.setEnabled(true);
		cmdCompute.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent e) {
				if (cboLeftOperand.getModel().getSize() > 0 && cboRightOperand.getModel().getSize() > 0){
					try {
						// Subtract left operand by right operand:
						Operation result = new Operation("$"+(cboLeftOperand.getSelectedIndex()+1)+"-$"+(cboRightOperand.getSelectedIndex()+1), plugin);
						SED resultSp = result.getSED();
						// Set its name, area and origin:
						if (resultSp != null && !resultSp.isEmpty()){
							resultSp.setName(txtResultName.getText());
							// Add it to the spectrums list of plugin:
							plugin.addSEDSource(result);
							plugin.showSED(resultSp);
							numSubtraction++;
						}
					} catch (SEDException ce) {
						setVisible(false);
						JOptionPane.showMessageDialog(BackgroundSubtractorDialog.this, ce.getMessage(), ce.getTitle(), JOptionPane.ERROR_MESSAGE);
					}
					// Close the background subtractor dialog:
					dispose();
				}
			}
		});

		cmdCancel = new JButton("Cancel");
		cmdCancel.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent e) {
				dispose();
			}
		});

		JPanel panCommands = new JPanel();
		panCommands.add(cmdCompute);
		panCommands.add(cmdCancel);

		// GENERAL:
		// Content:
		getContentPane().setLayout(new BorderLayout());
		getContentPane().add(panResultName, BorderLayout.NORTH);
		getContentPane().add(panOperation, BorderLayout.CENTER);
		getContentPane().add(panCommands, BorderLayout.SOUTH);
		// Title:
		setTitle("Background subtractor");
		// Other:
		pack();
		URL iconTitle = getClass().getResource(SEDPlugin.ICONS_DIR+"calculator.gif");
		if (iconTitle != null)
			setIconImage(new ImageIcon(iconTitle).getImage());
		setModal(true);
		setResizable(false);
		setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
	}

	private final void buildSEDIcon(){
		// Load all required icons:
		if (icoPos == null){
			URL urlIcon = getClass().getResource(SEDPlugin.ICONS_DIR+"position.gif");
			if (urlIcon != null)
				icoPos = new ImageIcon(urlIcon);
		}
		if (icoCircle == null){
			URL urlIcon = getClass().getResource(SEDPlugin.ICONS_DIR+"circle.gif");
			if (urlIcon != null)
				icoCircle = new ImageIcon(urlIcon);
		}
		if (icoPolygon == null){
			URL urlIcon = getClass().getResource(SEDPlugin.ICONS_DIR+"polygon.gif");
			if (urlIcon != null)
				icoPolygon = new ImageIcon(urlIcon);
		}
		if (icoOperation == null){
			URL urlIcon = getClass().getResource(SEDPlugin.ICONS_DIR+"calculator.gif");
			if (urlIcon != null)
				icoOperation = new ImageIcon(urlIcon);
		}
		if (icoCatalog == null){
			URL urlIcon = getClass().getResource(SEDPlugin.ICONS_DIR+"source3.png");
			if (urlIcon != null)
				icoCatalog = new ImageIcon(urlIcon);
		}
	}

	/* ************* */
	/* OTHER METHODS */
	/* ************* */
	public static final void reset(){
		numSubtraction = 0;
	}

	/* ****************** */
	/* LIST CELL RENDERER */
	/* ****************** */
	protected final class SEDListCellRenderer extends JLabel implements ListCellRenderer {
		private static final long serialVersionUID = 1L;

		public Component getListCellRendererComponent(JList list, Object value, int index, boolean isSelected, boolean cellHasFocus) {
			SED sed = (SED)value;
			Color color = plugin.getColor(sed);

			// set the SED name:
			setText(sed.getName());

			// set an icon according to the spectrum origin:
			if (sed.getSource() != null){
				SEDSource src = sed.getSource();
				String srcType = src.getType();
				if (srcType.equals(Circle.CIRCLE_TYPE)){
					setIcon(icoCircle);
					color = ((Circle)src).getColor();
				}else if (srcType.equals(Polygon.POLYGON_TYPE)){
					setIcon(icoPolygon);
					color = ((Polygon)src).getColor();
				}else if (srcType.equals(Operation.OPERATION_TYPE))
					setIcon(icoOperation);
				else
					setIcon(null);
			}

			// set the text color:
			setForeground(color);

			// set the background according the focus of the item:
			if (cellHasFocus)
				setBackground(list.getSelectionBackground());
			else
				setBackground(list.getBackground());

			return this;
		}

	}

}
