package sedplugin.sed;

import java.beans.PropertyChangeListener;

import java.text.SimpleDateFormat;

import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.Iterator;
import java.util.Vector;

import sedplugin.sed.source.SEDSource;

import sedplugin.sed.unit.XAxis;
import sedplugin.sed.unit.YAxis;

/**
 * Representation of a SED (Spectral Energy Distribution).
 * 
 * @author Gr&eacute;gory Mantelet (CDS)
 * @version 04/2011
 * 
 * @see SEDPoint
 */
public class SED implements Iterable<SEDPoint>, Exportable {

	/** SED name/label. */
	private String name;

	/** List of all points which compose the spectrum. */
	private final Vector<SEDPoint> points;

	/** Lets sorting the points. */
	private final PointComparator comparator;

	/** A tool object which manages units along the abscissa axis. */
	private final XAxis xUnit;

	/** A tool object which manages units along the ordinate axis. */
	private final YAxis yUnit;

	/** Abscissa value limits.
	 * @see Spectrum#computeExtremas() */
	protected double[] xExtremas;

	/** Ordinate value limits.
	 * @see Spectrum#computeExtremas() */
	protected double[] yExtremas;

	/** Source or origin of this SED. */
	private SEDSource source = null;

	/** Objects to notify when this SED has just been updated. */
	private final Vector<SEDListener> updateListeners = new Vector<SEDListener>();

	/** Objects to notify of the loading progression of this SED. */
	private final Vector<PropertyChangeListener> loadListeners = new Vector<PropertyChangeListener>();

	/* ************ */
	/* CONSTRUCTORS */
	/* ************ */
	/**
	 * Builds a SED with the given units.
	 * 
	 * @param xUnit	Unit of the abscissa axis.
	 * @param yUnit	Unit of the ordinate axis.
	 * 
	 * @see #SED(String, XAxis, YAxis, SEDSource)
	 */
	public SED(XAxis xUnit, YAxis yUnit){
		this("", xUnit, yUnit, null);
	}

	/**
	 * Builds a SED with the given units and its source.
	 * 
	 * @param xUnit		Unit of the abscissa axis.
	 * @param yUnit		Unit of the ordinate axis.
	 * @param source	Source of this SED.
	 * 
	 * @see #SED(String, XAxis, YAxis, SEDSource)
	 */
	public SED(XAxis xUnit, YAxis yUnit, SEDSource source){
		this("", xUnit, yUnit, source);
	}

	/**
	 * Builds a SED with its name, the given units and its source.
	 * 
	 * @param name		SED name.
	 * @param xUnit		Unit of the abscissa axis.
	 * @param yUnit		Unit of the ordinate axis.
	 * @param source	Source of this SED.
	 */
	public SED(String name, XAxis xUnit, YAxis yUnit, SEDSource source){
		this.name = (name==null)?"":name.trim();

		this.xUnit = xUnit;
		this.xUnit.addSED(this);
		this.yUnit = yUnit;
		this.yUnit.addSED(this);

		this.source = source;

		points = new Vector<SEDPoint>(4, 10);
		comparator = new PointComparator(PointComparator.ABSCISSA_ORDER);
		xExtremas = new double[]{Double.NaN, Double.NaN, Double.NaN};
		yExtremas = new double[]{Double.NaN, Double.NaN, Double.NaN};
	}

	/* ******************* */
	/* EXTREMAS MANAGEMENT */
	/* ******************* */
	/**
	 * Explores the whole spectrum to get its abscissa and ordinate extremas.
	 */
	protected void computeExtremas(){
		xExtremas = new double[]{Double.POSITIVE_INFINITY, Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY};
		yExtremas = new double[]{Double.POSITIVE_INFINITY, Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY};

		for(Iterator<SEDPoint> it=points.iterator(); it.hasNext();){
			SEDPoint sp = it.next();
			double x=sp.getX(), y=sp.getY();

			if (x < xExtremas[0]) xExtremas[0] = x;
			if (x > xExtremas[1]) xExtremas[1] = x;
			if (x > 0 && x < xExtremas[2]) xExtremas[2] = x;

			if (y < yExtremas[0]) yExtremas[0] = y;
			if (y > yExtremas[1]) yExtremas[1] = y;
			if (y > 0 && y < yExtremas[2]) yExtremas[2] = y;
		}
	}

	/* ***************** */
	/* GETTERS & SETTERS */
	/* ***************** */
	/**
	 * Indicates the way the points are sorted/ordered.
	 * 
	 * @return The pointOrder ({@link PointComparator#NO_ORDER}, {@link PointComparator#ABSCISSA_ORDER} or {@link PointComparator#ORDINATE_ORDER}).
	 */
	public final int getPointOrder() {
		return comparator.getPointOrder();
	}

	/**
	 * <p>Lets specify in which order the point must be sorted.</p>
	 * 
	 * <p><i><u>Note:</u> The points are immediately re-ordered.</i></p>
	 * 
	 * @param pointOrder The pointOrder to set ({@link PointComparator#NO_ORDER}, {@link PointComparator#ABSCISSA_ORDER} or {@link PointComparator#ORDINATE_ORDER}).
	 */
	public final void setPointOrder(int order) {
		if (comparator.setPointOrder(order))
			Collections.sort(points, comparator);
	}

	/**
	 * Gets the name of this SED.
	 * 
	 * @return Its name.
	 */
	public final String getName() {
		return name;
	}

	/**
	 * Sets the name of this SED.
	 * 
	 * @param name Its new name.
	 */
	public final void setName(String name) {
		this.name = (name==null)?"":name.trim();
	}

	/**
	 * Gets the used abscissa unit.
	 * 
	 * @return Its X unit.
	 */
	public final XAxis getXAxis() {
		return xUnit;
	}

	/**
	 * Gets the used ordinate unit.
	 * 
	 * @return Its Y unit.
	 */
	public final YAxis getYAxis() {
		return yUnit;
	}

	/**
	 * Gets the object which has built this SED and which must update it.
	 * 
	 * @return	Its source.
	 */
	public final SEDSource getSource(){
		return source;
	}

	/**
	 * Sets the source of this SED.
	 * 
	 * @param src	Its source.
	 */
	public final void setSource(final SEDSource src){
		source = src;
	}

	/**
	 * Gets the minimum and the maximum values on the abscissa axis of this SED.
	 * 
	 * @return An array of two items: [0] = X minimum and [1] = X maximum.
	 */
	public final double[] getXExtremas() {
		return xExtremas;
	}

	/**
	 * Gets the minimum and the maximum values on the ordinate axis of this SED.
	 * 
	 * @return An array of two items: [0] = Y minimum and [1] = Y maximum.
	 */
	public final double[] getYExtremas() {
		return yExtremas;
	}


	/* ******************** */
	/* LISTENERS MANAGEMENT */
	/* ******************** */
	/**
	 * Adds an object which wants to be notify at each update of this SED.
	 * 
	 * @param l	The object to notify.
	 * @return	<i>true</i> if the listener has been successfully added, <i>false</i> otherwise (or if the object is already a listener of this SED).
	 */
	public final boolean addSEDListener(final SEDListener l){
		if (updateListeners.contains(l))
			return false;
		else
			return updateListeners.add(l);
	}

	/**
	 * Removes the given object from the list of objects to notify at each update of this SED.
	 * 
	 * @param l	The object to remove.
	 * @return	<i>true</i> if the given listener has been successfully removed, <i>false</i> otherwise.
	 */
	public final boolean removeSEDListener(final SEDListener l){
		return updateListeners.remove(l);
	}

	/**
	 * Notifies all listeners of this SED that it has just been updated.
	 * 
	 * @see SEDListener#sedUpdated(SED, SEDSource)
	 */
	public final void notifySEDListeners(){
		for(SEDListener l : updateListeners)
			l.sedUpdated(this, source);
	}

	/**
	 * Adds an object which wants to be notify of the progression of this SED.
	 * 
	 * @param l	The object to notify.
	 * @return	<i>true</i> if the listener has been successfully added, <i>false</i> otherwise (or if the object is already a listener of this SED).
	 */
	public final boolean addSEDLoadListener(final PropertyChangeListener l){
		if (loadListeners.contains(l))
			return false;
		else
			return loadListeners.add(l);
	}

	/**
	 * Removes the given object from the list of objects to notify of the progression of this SED.
	 * 
	 * @param l	The object to remove.
	 * @return	<i>true</i> if the given listener has been successfully removed, <i>false</i> otherwise.
	 */
	public final boolean removeSEDLoadListener(final PropertyChangeListener l){
		return loadListeners.remove(l);
	}

	/**
	 * Gets all objects to notify of the loading progression of this SED.
	 */
	public final Iterator<PropertyChangeListener> getLoadListeners(){
		return loadListeners.iterator();
	}


	/* ***************** */
	/* POINTS MANAGEMENT */
	/* ***************** */
	/**
	 * Adds the given point (if it is not already done) into this SED.
	 * 
	 * @param newPoint	The point to add.
	 * 
	 * @return	<i>true</i> if the given point has been added, <i>false</i> otherwise.
	 * 
	 * @see SEDPoint#setSed(SED)
	 */
	public final boolean add(SEDPoint newPoint){
		if (newPoint == null)
			return false;

		newPoint.setSed(this);
		if (orderedInsert(newPoint)){
			computeExtremas();
			return true;
		}else{
			newPoint.setSed(null);
			return false;
		}
	}

	/**
	 * Inserts the given point considering the current order of all the SED points.
	 * 
	 * @param newPoint	The point to add.
	 * 
	 * @return	<i>true</i> if the point has been added, <i>false</i> otherwise.
	 */
	protected final boolean orderedInsert(SEDPoint newPoint){
		if (newPoint == null)
			return false;

		if (points.isEmpty())
			return points.add(newPoint);

		int start=0, end=points.size()-1;

		while(start<end){
			int middle = (start+end)/2;
			int comp = comparator.compare(newPoint, points.get(middle));
			if (comp == 0)
				return false;
			else if (comp < 0)
				end = middle-1;
			else
				start = middle+1;
		}

		int comp = comparator.compare(newPoint, points.get(start));
		if (comp == 0)
			return false;
		else if (comp < 0)
			points.add(start, newPoint);
		else
			points.add(start+1, newPoint);

		return true;
	}

	/**
	 * Adds all the given points in this SED.
	 * 
	 * @param newPoints	The points to add.
	 * 
	 * @return	The number of added points.
	 */
	public final int addAll(SEDPoint[] newPoints){
		int nbAddedPoints = 0;

		for(int i=0; i<newPoints.length; i++){
			SEDPoint p = newPoints[i];
			p.setSed(this);
			if (orderedInsert(p))
				nbAddedPoints++;
			else
				p.setSed(null);
		}

		computeExtremas();

		return nbAddedPoints;
	}

	/**
	 * Adds all the given points.
	 * 
	 * @param newPoints	The points to add.
	 * 
	 * @return	The number of added points.
	 */
	public final int addAll(Collection<SEDPoint> newPoints){
		if (newPoints == null)
			return 0;

		int nbAddedPoints = 0;

		Iterator<SEDPoint> it = newPoints.iterator();
		while(it.hasNext()){
			SEDPoint p = it.next();
			p.setSed(this);
			if (orderedInsert(p))
				nbAddedPoints++;
			else
				p.setSed(null);
		}

		computeExtremas();

		return nbAddedPoints;
	}

	/**
	 * Gets the indPoint-th points of this SED (considering the order).
	 * 
	 * @param indPoint	The index of the point to get.
	 * 
	 * @return	The corresponding point.
	 */
	public final SEDPoint get(int indPoint){
		return (SEDPoint)points.get(indPoint);
	}

	/**
	 * Gets all the points of this SED.
	 * 
	 * @return	An iterator on all its points.
	 */
	public final Iterator<SEDPoint> getPoints(){
		return points.iterator();
	}

	/**
	 * Gets all the points of this SED.
	 * 
	 * @return	An iterator on all its points.
	 * 
	 * @see #getPoints()
	 */
	public final Iterator<SEDPoint> iterator(){
		return getPoints();
	}

	/**
	 * Gets the number of all points contained in this SED.
	 * 
	 * @return	The number of its points.
	 */
	public final int size(){
		return points.size();
	}

	/**
	 * Indicates whether this SED is empty (no point) or not.
	 * 
	 * @return	<i>true</i> if there is no point in this SED, <i>false</i> otherwise.
	 */
	public final boolean isEmpty(){
		return points.isEmpty();
	}

	/**
	 * Removes the specified point from this SED.
	 * 
	 * @param indPoint	The index of the point to remove.
	 * 
	 * @return	The removed point.
	 * 
	 * @see SEDPoint#setSed(SED)
	 */
	public final SEDPoint remove(int indPoint){
		SEDPoint removed = points.remove(indPoint);
		if (removed != null){
			removed.setSed(null);
			computeExtremas();
			return removed;
		}
		return null;
	}

	/**
	 * Removes all the points of this SED.
	 */
	public final void clear(){
		for(int i=points.size()-1; i>=0; i--){
			SEDPoint removed = points.remove(i);
			if (removed != null)
				removed.setSed(null);
		}
		computeExtremas();
	}

	/* ******************* */
	/* UNIT UPDATE METHODS */
	/* ******************* */
	/**
	 * Updates the X value of all the points of this SED.
	 * 
	 * @see SEDPoint#updateX()
	 */
	public void updateX(){
		Iterator<SEDPoint> it = points.iterator();
		while(it.hasNext())
			it.next().updateX();

		if (comparator != null && comparator.getPointOrder() == PointComparator.ABSCISSA_ORDER)
			Collections.sort(points, comparator);

		computeExtremas();
	}

	/**
	 * Update the Y value of all the points of this SED.
	 * 
	 * @see SEDPoint#updateY()
	 */
	public void updateY(){
		Iterator<SEDPoint> it = points.iterator();
		while(it.hasNext())
			it.next().updateY();

		if (comparator != null && comparator.getPointOrder() == PointComparator.ORDINATE_ORDER)
			Collections.sort(points, comparator);

		computeExtremas();
	}


	/* ***************** */
	/* INHERITED METHODS */
	/* ***************** */
	@Override
	public boolean equals(Object obj){
		if (obj instanceof SED){
			SED sed2 = (SED)obj;
			return name.equals(sed2.name) && points.equals(sed2.points);
		}
		return false;
	}

	@Override
	public String toString(){
		/*String str = name;
		Iterator<SEDPoint> it = points.iterator();
		while(it.hasNext())
			str += "\n\t* "+it.next().toString();
		return str;*/
		return getName();
	}

	/* **************** */
	/* EXPORT FUNCTIONS */
	/* **************** */
	public String toASCIIFormat(String sep) {
		String txt = "";

		if (name != null && name.trim().length()>0)
			txt += "\n%% SED - "+getName()+"\n";
		txt += "% "+xUnit.getUnit().symbol+sep+yUnit.getUnit().symbol+"\n";

		for(Iterator<SEDPoint> it = points.iterator(); it.hasNext();)
			txt += it.next().toASCIIFormat(sep)+"\n";

		return txt;
	}

	public String toVOTableFormat(String linePrefix){
		SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd'T'kk:mm:ss'Z'");

		if (linePrefix == null)
			linePrefix = "";

		StringBuffer voTable = new StringBuffer();
		voTable.append(linePrefix).append("<GROUP utype=\"spec:Char\">\n");

		// CHARACTERIZATION[SPATIAL_AXIS] -> COMPLETE:
		String sourceVotable = source.toVOTableFormat(linePrefix+"\t");
		if (sourceVotable != null)
			voTable.append(sourceVotable);

		// CHARACTERIZATION[TIME_AXIS] -> To Complete:
		StringBuffer charTimeAxis = new StringBuffer(linePrefix);
		charTimeAxis.append("\t<GROUP utype=\"spec:Char.TimeAxis\">\n");
		charTimeAxis.append(linePrefix).append("\t\t<GROUP utype=\"spec:Char.TimeAxis.Coverage\">\n");
		charTimeAxis.append(linePrefix).append("\t\t\t<GROUP utype=\"spec:Char.TimeAxis.Coverage.Location\">\n");
		charTimeAxis.append(linePrefix).append("\t\t\t\t<PARAM name=\"TimeObs\" utype=\"spec:Char.TimeAxis.Coverage.Value\" ucd=\"time.obs\" datatype=\"double\" unit=\"d\" value=\"null\" />\n");
		charTimeAxis.append(linePrefix).append("\t\t\t</GROUP>\n");

		// CHARACTERIZATION[SPECTRAL_AXIS] -> To Complete:
		StringBuffer charSpectralAxis = new StringBuffer(linePrefix);
		charSpectralAxis.append("\t<GROUP utype=\"spec:Char.SpectralAxis\">\n");
		charSpectralAxis.append(linePrefix).append("\t\t<GROUP utype=\"spec:Char.SpectralAxis.Coverage\">\n");
		charSpectralAxis.append(linePrefix).append("\t\t\t<GROUP utype=\"spec:Char.SpectralAxis.Coverage.Location\">\n");
		charSpectralAxis.append(linePrefix).append("\t\t\t\t<PARAM name=\"SpectralLocation\" utype=\"spec:Char.SpectralAxis.Coverage.Location.Value\" ucd=\"em.wl.central;instr.bandpass\" datatype=\"double\" value=\"null\" />\n");
		charSpectralAxis.append(linePrefix).append("\t\t\t</GROUP>\n");

		// CURATION -> COMPLETE:
		StringBuffer voTableEnd = new StringBuffer(linePrefix);
		voTableEnd.append("<GROUP utype=\"spec:Curation\">\n");
		voTableEnd.append(linePrefix).append("\t<PARAM name=\"Publisher\" utype=\"spec:Curation.Publisher\" ucd=\"meta.curation\" datatype=\"char\" arraysize=\"*\" value=\"CDS - Aladin/SEDPlugin\" />\n");
		voTableEnd.append(linePrefix).append("</GROUP>\n");

		// DATA_ID -> COMPLETE:
		voTableEnd.append(linePrefix).append("<GROUP utype=\"spec:DataID\">\n");
		voTableEnd.append(linePrefix).append("\t<PARAM name=\"Title\" utype=\"spec:DataID.Title\" ucd=\"meta.title\" datatype=\"char\" arraysize=\"*\" value=\"").append(name).append("\" />\n");
		voTableEnd.append(linePrefix).append("\t<PARAM name=\"Creator\" utype=\"spec:DataID.Creator\" datatype=\"char\" arraysize=\"*\" value=\"CDS - Aladin/SEDPlugin").append(sedplugin.plugin.SEDPlugin.VERSION).append("\" />\n");
		voTableEnd.append(linePrefix).append("\t<PARAM name=\"DataDate\" utype=\"spec:DataID.Date\" ucd=\"time;meta.dataset\" datatype=\"char\" arraysize=\"*\" value=\"").append(sdf.format(new Date())).append("\" />\n");
		voTableEnd.append(linePrefix).append("</GROUP>\n");

		// spec:DATA -> COMPLETE:
		voTableEnd.append(linePrefix).append("<GROUP utype=\"spec:Data\">\n");
		voTableEnd.append(linePrefix).append("\t<GROUP utype=\"spec:Data.SpectralAxis\">\n");
		voTableEnd.append(linePrefix).append("\t\t<FIELDref ref=\"SpectralValue0\" />\n");
		voTableEnd.append(linePrefix).append("\t</GROUP>\n");
		voTableEnd.append(linePrefix).append("\t<GROUP utype=\"spec:Data.FluxAxis\">\n");
		voTableEnd.append(linePrefix).append("\t\t<FIELDref ref=\"FluxValue0\" />\n");
		voTableEnd.append(linePrefix).append("\t</GROUP>\n");
		voTableEnd.append(linePrefix).append("</GROUP>\n");

		// FIELDS -> COMPLETE:
		boolean isWavelength = xUnit.isQuantity(XAxis.WAVELENGTH);
		voTableEnd.append(linePrefix).append("<FIELD name=\"").append(xUnit.getQuantity().getSymbol()).append("\" ID=\"SpectralValue0\" utype=\"spec:Data.SpectralAxis.Value\" ucd=\"").append((isWavelength)?"em.wl":"em.freq").append("\" datatype=\"double\" unit=\"").append(xUnit.getUnit().symbol).append("\" />\n");
		voTableEnd.append(linePrefix).append("<FIELD name=\"Flux\" ID=\"FluxValue0\" utype=\"spec:Data.FluxAxis.Value\" ucd=\"phot.flux.density\" datatype=\"double\" unit=\"").append(yUnit.getUnit().symbol).append("\" />\n");

		// DATA -> To Complete:
		voTableEnd.append(linePrefix).append("<DATA>\n").append(linePrefix).append("\t<TABLEDATA>\n");

		// Fill DATA:
		double minSpectral=Double.MAX_VALUE, maxSpectral=Double.MIN_VALUE;
		double minDate=Double.MAX_VALUE, maxDate=Double.MIN_VALUE;
		for(Iterator<SEDPoint> it=points.iterator(); it.hasNext();){
			SEDPoint spt = it.next();
			double wl=spt.getX(), date=Double.NaN;
			if (spt.getImage() != null){
				try{
					date = Double.parseDouble(spt.getImage().getCharacterization().getDate());
				}catch(NumberFormatException nfe){;}
			}
			// get its spectral value:
			minSpectral = Math.min(minSpectral, wl);
			maxSpectral = Math.max(maxSpectral, wl);
			// get its date:
			minDate = Math.min(minDate, date);
			maxDate = Math.max(maxDate, date);
			// add a new row:
			voTableEnd.append(spt.toVOTableFormat(linePrefix+"\t\t"));
		}

		// Finish DATA:
		voTableEnd.append(linePrefix).append("\t</TABLEDATA>\n").append(linePrefix).append("</DATA>\n");

		// Complete CHARACTERIZATION[TIME_AXIS]:
		charTimeAxis.append(linePrefix).append("\t\t\t<GROUP utype=\"spec:Char.TimeAxis.Coverage.Bounds\">\n");
		charTimeAxis.append(linePrefix).append("\t\t\t\t<PARAM name=\"TimeExtent\" utype=\"spec:Char.TimeAxis.Coverage.Bounds.Extent\" ucd=\"time.duration\" datatype=\"double\" unit=\"d\" value=\"").append(maxDate-minDate).append("\" />\n");
		charTimeAxis.append(linePrefix).append("\t\t\t\t<PARAM name=\"MinTime\" utype=\"spec:Char.TimeAxis.Coverage.Bounds.Start\" ucd=\"time.epoch\" datatype=\"double\" unit=\"d\" value=\"").append(minDate).append("\" />\n");
		charTimeAxis.append(linePrefix).append("\t\t\t\t<PARAM name=\"MaxTime\" utype=\"spec:Char.TimeAxis.Coverage.Bounds.Stop\" ucd=\"time.epoch\" datatype=\"double\" unit=\"d\" value=\"").append(maxDate).append("\" />\n");
		charTimeAxis.append(linePrefix).append("\t\t\t</GROUP>\n");
		charTimeAxis.append(linePrefix).append("\t\t</GROUP>\n");
		charTimeAxis.append(linePrefix).append("\t</GROUP>\n");

		// Complete CHARACTERIZATION[SPECTRAL_AXIS]:
		charSpectralAxis.append(linePrefix).append("\t\t\t<GROUP utype=\"spec:Char.SpectralAxis.Coverage.Bounds\">\n");
		charSpectralAxis.append(linePrefix).append("\t\t\t\t<PARAM name=\"MinSpectral\" utype=\"spec:Char.SpectralAxis.Coverage.Bounds.Start\" ucd=\"stat.min\" datatype=\"double\" unit=\"").append(xUnit.getUnit().symbol).append("\" value=\"").append(minSpectral).append("\" />\n");
		charSpectralAxis.append(linePrefix).append("\t\t\t\t<PARAM name=\"MaxSpectral\" utype=\"spec:Char.SpectralAxis.Coverage.Bounds.Stop\" ucd=\"stat.max\" datatype=\"double\" unit=\"").append(xUnit.getUnit().symbol).append("\" value=\"").append(maxSpectral).append("\" />\n");
		charSpectralAxis.append(linePrefix).append("\t\t\t</GROUP>\n");
		charSpectralAxis.append(linePrefix).append("\t\t</GROUP>\n");
		charSpectralAxis.append(linePrefix).append("\t</GROUP>\n");

		// Concatenate all these blocks:
		voTable.append(charTimeAxis).append(charSpectralAxis);
		voTable.append(linePrefix).append("</GROUP>\n");
		voTable.append(voTableEnd);

		return voTable.toString();
	}

	public String toXMLFormat(String linePrefix){
		SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd'T'kk:mm:ss'Z'");

		if (linePrefix == null)
			linePrefix = "";

		StringBuffer xml = new StringBuffer();
		xml.append(linePrefix).append("<Char>\n");

		// CHARACTERIZATION[SPATIAL_AXIS] -> COMPLETE:
		String sourceXml = source.toXMLFormat(linePrefix+"\t");
		if (sourceXml != null)
			xml.append(sourceXml);

		// CHARACTERIZATION[TIME_AXIS] -> To Complete:
		StringBuffer charTimeAxis = new StringBuffer();
		charTimeAxis.append(linePrefix).append("\t<TimeAxis name=\"Time\" ucd=\"time\" unit=\"d\">\n");
		charTimeAxis.append(linePrefix).append("\t\t<CoordSystem idref=\"ID000001\" />\n");
		charTimeAxis.append(linePrefix).append("\t\t<Coverage>\n");
		charTimeAxis.append(linePrefix).append("\t\t\t<Location>\n");
		charTimeAxis.append(linePrefix).append("\t\t\t\t<Value ucd=\"time.obs\" unit=\"d\">null</Value>\n");
		charTimeAxis.append(linePrefix).append("\t\t\t</Location>\n");

		// CHARACTERIZATION[SPECTRAL_AXIS] -> To Complete:
		StringBuffer charSpectralAxis = new StringBuffer();
		charSpectralAxis.append(linePrefix).append("\t<SpectralAxis name=\"SpectralCoord\" ucd=\"em.wl\" unit=\"").append(xUnit.getUnit().symbol).append("\">\n");
		charSpectralAxis.append(linePrefix).append("\t\t<CoordSystem idref=\"ID000001\" />\n");
		charSpectralAxis.append(linePrefix).append("\t\t<Coverage>\n");
		charSpectralAxis.append(linePrefix).append("\t\t\t<Location>\n");
		charSpectralAxis.append(linePrefix).append("\t\t\t\t<Value ucd=\"em.wl.central;instr.bandpass\">null</Value>\n");
		charSpectralAxis.append(linePrefix).append("\t\t\t</Location>\n");

		// CHARACTERIZATION[FLUX_AXIS] -> To Complete:
		StringBuffer charFluxAxis = new StringBuffer();
		charFluxAxis.append(linePrefix).append("\t<FluxAxis name=\"FluxCoord\" ucd=\"phot.flux.density\" unit=\"").append(yUnit.getUnit().symbol).append("\">\n");
		charFluxAxis.append(linePrefix).append("\t</FluxAxis>\n");

		// CURATION -> COMPLETE:
		StringBuffer xmlEnd = new StringBuffer();
		xmlEnd.append(linePrefix).append("<Curation>\n");
		xmlEnd.append(linePrefix).append("\t<Publisher ucd=\"meta.curation\">CDS - Aladin/SEDPlugin</Publisher>\n");
		xmlEnd.append(linePrefix).append("</Curation>\n");

		// DATA_ID -> COMPLETE:
		xmlEnd.append(linePrefix).append("<DataID>\n");
		xmlEnd.append(linePrefix).append("\t<Title ucd=\"meta.title\">").append(name).append("</Title>\n");
		xmlEnd.append(linePrefix).append("\t<Creator>CDS - Aladin/SEDPlugin ").append(sedplugin.plugin.SEDPlugin.VERSION).append("</Creator>\n");
		xmlEnd.append(linePrefix).append("\t<Date ucd=\"time;meta.dataset\">").append(sdf.format(new Date())).append("</Date>\n");
		xmlEnd.append(linePrefix).append("</DataID>\n");

		// DATA -> To Complete:
		xmlEnd.append(linePrefix).append("<ArrayOfPoint>\n");

		// Fill DATA:
		double minSpectral=Double.MAX_VALUE, maxSpectral=Double.MIN_VALUE;
		double minDate=Double.MAX_VALUE, maxDate=Double.MIN_VALUE;
		for(Iterator<SEDPoint> it=points.iterator(); it.hasNext();){
			SEDPoint spt = it.next();
			double wl=spt.getX(), date=Double.NaN;
			if (spt.getImage() != null){
				try{
					date = Double.parseDouble(spt.getImage().getCharacterization().getDate());
				}catch(NumberFormatException nfe){;}
			}
			// get its spectral value:
			minSpectral = Math.min(minSpectral, wl);
			maxSpectral = Math.max(maxSpectral, wl);
			// get its date:
			minDate = Math.min(minDate, date);
			maxDate = Math.max(maxDate, date);
			// add a new row:
			xmlEnd.append(spt.toXMLFormat(linePrefix+"\t"));
		}

		// Finish DATA:
		xmlEnd.append(linePrefix).append("</ArrayOfPoint>\n");

		// Complete CHARACTERIZATION[TIME_AXIS]:
		charTimeAxis.append(linePrefix).append("\t\t\t<Bounds>\n");
		charTimeAxis.append(linePrefix).append("\t\t\t\t<Extent ucd=\"time.duration\" unit=\"d\">").append(maxDate-minDate).append("</Extent>\n");
		charTimeAxis.append(linePrefix).append("\t\t\t\t<Start ucd=\"time.epoch\" unit=\"d\">").append(minDate).append("</Start>\n");
		charTimeAxis.append(linePrefix).append("\t\t\t\t<Stop ucd=\"time.epoch\" unit=\"d\">").append(maxDate).append("</Stop>\n");
		charTimeAxis.append(linePrefix).append("\t\t\t</Bounds>\n");
		charTimeAxis.append(linePrefix).append("\t\t</Coverage>\n");
		charTimeAxis.append(linePrefix).append("\t</TimeAxis>\n");

		// Complete CHARACTERIZATION[SPECTRAL_AXIS]:
		charSpectralAxis.append(linePrefix).append("\t\t\t<Bounds>\n");
		charSpectralAxis.append(linePrefix).append("\t\t\t\t<Start ucd=\"stat.min\" unit=\"").append(xUnit.getUnit().symbol).append("\">").append(minSpectral).append("</Start>\n");
		charSpectralAxis.append(linePrefix).append("\t\t\t\t<Stop ucd=\"stat.max\" unit=\"").append(xUnit.getUnit().symbol).append("\">").append(maxSpectral).append("</Stop>\n");
		charSpectralAxis.append(linePrefix).append("\t\t\t</Bounds>\n");
		charSpectralAxis.append(linePrefix).append("\t\t</Coverage>\n");
		charSpectralAxis.append(linePrefix).append("\t</SpectralAxis>\n");

		// Concatenate all these blocks:
		xml.append(charTimeAxis).append(charSpectralAxis).append(charFluxAxis);
		xml.append(linePrefix).append("</Char>\n");
		xml.append(xmlEnd);

		return xml.toString();
	}

	/* ********* */
	/* SELF TEST */
	/* ********* */
	public static boolean trace = false;

	public static void main(String[] args) throws Exception {
		XAxis xAxis = new XAxis();
		YAxis yAxis = new YAxis(xAxis);

		if (args.length < 5){
			System.err.println("((X)) Missing "+(5-args.length)+" required parameters ! ((X))");
			System.err.println("Usage: sed [-debug] x xunit y yunit surface");
			System.exit(1);
		}

		int indParam = 0;

		if (args[indParam].equalsIgnoreCase("-debug")){
			trace = true;
			indParam++;
		}
		try{
			double x = Double.parseDouble(args[indParam++]);
			String xUnit = args[indParam++];
			double y = Double.parseDouble(args[indParam++]);
			String yUnit = args[indParam++];
			double surface = Double.parseDouble(args[indParam++]);

			if (!xAxis.unitExists(xUnit)){
				System.err.println("((X)) ERROR: "+xUnit+" is not a known abscissa unit ! ((X))");
				System.exit(2);
			}
			if (!yAxis.unitExists(yUnit)){
				System.err.println("((X)) ERROR: "+yUnit+" is not a known ordinate unit ! ((X))");
				System.exit(3);
			}

			SEDPoint point = new SEDPoint(x, xUnit, y, yUnit, surface);

			SED sed = new SED("TestSED", xAxis, yAxis, null);
			sed.add(point);

			System.out.println("##### INITIAL VALUES #####");
			System.out.println("( "+ToolBox.format(point.getOriginalX())+point.getOriginalXUnit()+" ; "+ToolBox.format(point.getOriginalY())+point.getOriginalYUnit()+" )\n");

			System.out.println("##### CONVERSIONS #####");
			for(int i=0; i<yAxis.getNbQuantities(); i++){
				yAxis.setQuantity(i);
				int nbUnits = yAxis.getQuantity().getNbUnits();
				for(int j=0; j<nbUnits; j++){
					yAxis.setUnit(j);
					System.out.println(yAxis+" => "+ToolBox.format(point.getY())+point.getYUnit());
				}
				System.out.println();
			}
		}catch(NumberFormatException nfe){
			indParam--;
			System.err.println("((X)) ERROR: the "+(indParam+1)+"th parameter ("+args[indParam]+") must be numeric !");
		}
	}

}
