package sedplugin.sed;

import sedplugin.errors.ErrorsLog;
import sedplugin.errors.SEDException;

import sedplugin.plugin.AladinImage;

/**
 * Representation of one point of a SED.
 * Its current units are the ones of its SED. If this point is not associated to a SED, its current X and Y values and units
 * are its original values and units.
 * 
 * @author Gr&eacute;gory Mantelet (CDS)
 * @version 03/2011
 * 
 * @see SED
 */
public class SEDPoint implements Exportable {

	/** Its original abscissa value. */
	protected final double originalX;

	/** Its original abscissa unit. */
	protected final String originalXUnit;

	/** Its original ordinate value. */
	protected final double originalY;

	/** Its original ordinate unit. */
	protected final String originalYUnit;

	/** Surface of the selected area for this image (in ''²). */
	protected double surface;

	/** Its SED. */
	protected SED sed = null;

	/** Its current abscissa value. */
	protected double currentX;

	/** Its current ordinate value. */
	protected double currentY;

	protected AladinImage image = null;

	/* ************ */
	/* CONSTRUCTORS */
	/* ************ */
	/**
	 * Builds a SED point with its original values.
	 * 
	 * @param x			Its original X value.
	 * @param xUnit		Its original X unit.
	 * @param y			Its original Y value.
	 * @param yUnit		Its original Y unit.
	 * 
	 * @see #SEDPoint(double, String, double, String, double)
	 */
	public SEDPoint(double x, String xUnit, double y, String yUnit){
		this(x, xUnit, y, yUnit, 1);
	}

	/**
	 * Builds a SED point with its original values + its surface (in ''² = arcsec²).
	 * 
	 * @param x			Its original X value.
	 * @param xUnit		Its original X unit.
	 * @param y			Its original Y value.
	 * @param yUnit		Its original Y unit.
	 * @param surface	The surface of its source (in arcsec²).
	 */
	public SEDPoint(double x, String xUnit, double y, String yUnit, double surface){
		originalX = x;
		originalXUnit = xUnit;
		originalY = y;
		originalYUnit = yUnit;
		this.surface = surface;

		currentX = originalX;
		currentY = originalY;
	}

	/**
	 * Builds a SED point with its original values (extracted from the Characterization of the given image) + its surface( in''² = arcsec²).
	 * 
	 * @param image		Image from which this SED point has been extracted.
	 * @param y			The ordinate value of this SED point.
	 * @param surface	The surface of its source (in arcsec²).
	 */
	public SEDPoint(AladinImage image, double y, double surface){
		this(image.getCharacterization().getWavelength(), image.getCharacterization().getWavelengthUnit(), y, image.getCharacterization().getFluxUnit(), surface);
		this.image = image;
	}

	/**
	 * Builds a copy of the given SEDPoint.
	 * Only the SED reference is not preserved.
	 * 
	 * @param pt	The SED point to copy.
	 */
	public SEDPoint(final SEDPoint pt){
		sed = null;
		originalX = pt.originalX;
		originalXUnit = pt.originalXUnit;
		originalY = pt.originalY;
		originalYUnit = pt.originalYUnit;
		surface = pt.surface;
		image = pt.image;
		currentX = pt.originalX;
		currentY = pt.originalY;
	}

	/* ***************** */
	/* GETTERS & SETTERS */
	/* ***************** */
	/**
	 * Gets its original abscissa value.
	 * 
	 * @return Its original X value.
	 */
	public final double getOriginalX() {
		return originalX;
	}

	/**
	 * Gets its original ordinate value.
	 * 
	 * @return Its original X unit.
	 */
	public final String getOriginalXUnit() {
		return originalXUnit;
	}

	/**
	 * Gets its original ordinate value.
	 * 
	 * @return Its original Y value.
	 */
	public final double getOriginalY() {
		return originalY;
	}

	/**
	 * Gets its original ordinate unit.
	 * 
	 * @return Its original Y unit.
	 */
	public final String getOriginalYUnit() {
		return originalYUnit;
	}

	/**
	 * Gets the surface of the source of this point.
	 * 
	 * @return The surface of its source (in ''²).
	 */
	public final double getSurface() {
		return surface;
	}

	/**
	 * Gets the SED in which this point is a part of.
	 * 
	 * @return Its SED.
	 */
	public final SED getSed() {
		return sed;
	}

	/**
	 * Sets the SED with which this point must interact.
	 * 
	 * @param sed Its new SED.
	 * 
	 * @see #updateX()
	 * @see #updateY()
	 */
	protected final void setSed(SED sed) {
		this.sed = sed;
		updateX();
		updateY();
	}

	/**
	 * Gets the abscissa value of this point.
	 * 
	 * @return Its current X value.
	 */
	public final double getX() {
		return currentX;
	}

	/**
	 * Gets the abscissa unit of this point.
	 * 
	 * @return	Its current X unit.
	 * 
	 * @see SED#getXAxis()
	 */
	public final String getXUnit(){
		return (sed==null)?getOriginalXUnit():sed.getXAxis().getUnit().symbol;
	}

	/**
	 * Gets the ordinate value of this point.
	 * 
	 * @return Its current Y value.
	 */
	public final double getY() {
		return currentY;
	}

	/**
	 * Gets the current Y unit of this point.
	 * 
	 * @return	Its current Y unit.
	 * 
	 * @see SED#getYAxis()
	 */
	public final String getYUnit(){
		return (sed==null)?getOriginalYUnit():sed.getYAxis().getUnit().symbol;
	}

	/**
	 * Gets the image linked to this SED point.
	 * 
	 * @return The linked image.
	 */
	public final AladinImage getImage() {
		return image;
	}

	/**
	 * Sets the image linked to this SED point.
	 * 
	 * @param image The linked image.
	 */
	public final void setImage(AladinImage image) {
		this.image = image;
	}

	/* ******************* */
	/* UNIT UPDATE METHODS */
	/* ******************* */
	/**
	 * Update the current X value of this point according to the current X unit of its SED.
	 * 
	 * @see SED#getXAxis()
	 * @see XAxis#convert(SEDPoint)
	 */
	public void updateX(){
		if (sed == null){
			currentX = originalX;
		}else{
			try {
				currentX = sed.getXAxis().convert(this);
			} catch (SEDException se) {
				currentX = Double.NaN;
				ErrorsLog.getCurrentInstance().addException(se);
			}
		}
	}

	/**
	 * Update the current Y value of this point according to the current Y unit of its SED.
	 * 
	 * @see SED#getYAxis()
	 * @see YAxis#convert(SEDPoint)
	 */
	public void updateY(){
		if (sed == null){
			currentY = originalY;
		}else{
			try {
				currentY = sed.getYAxis().convert(this);
			} catch (SEDException se) {
				currentY = Double.NaN;
				ErrorsLog.getCurrentInstance().addException(se);
			}
		}
	}

	/* ***************** */
	/* INHERITED METHODS */
	/* ***************** */
	/**
	 * Two SED points are equals only if all their original values and units are equals.
	 * 
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj){
		if (obj instanceof SEDPoint){
			SEDPoint p2 = (SEDPoint)obj;
			return originalX == p2.originalX && originalXUnit.equalsIgnoreCase(p2.originalXUnit)
			&& originalY == p2.originalY && originalYUnit.equalsIgnoreCase(p2.originalYUnit)
			&& surface == p2.surface;
		}else
			return false;
	}

	/**
	 * <p><i><u>Example: </u> ( 2,14E5um ; 3,56Jy )</i></p>
	 * 
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString(){
		return "( "+ToolBox.format(getX())+getXUnit()+" ; "+ToolBox.format(getY())+getYUnit()+" )";
	}

	/* **************** */
	/* EXPORT FUNCTIONS */
	/* **************** */
	public String toASCIIFormat(String sep) {
		return getX()+sep+getY();
	}

	public String toVOTableFormat(String linePrefix){
		return ((linePrefix!=null)?linePrefix:"")+"<TR><TD>"+getX()+"</TD><TD>"+getY()+"</TD></TR>\n";
	}

	public String toXMLFormat(String linePrefix){
		StringBuffer xml = new StringBuffer();
		xml.append(linePrefix).append("<Point>\n");

		xml.append(linePrefix).append("\t<SpectralAxis>\n");
		xml.append(linePrefix).append("\t\t<Value>").append(getX()).append("</Value>\n");
		xml.append(linePrefix).append("\t</SpectralAxis>\n");
		xml.append(linePrefix).append("\t<FluxAxis>\n");
		xml.append(linePrefix).append("\t\t<Value>").append(getY()).append("</Value>\n");
		xml.append(linePrefix).append("\t</FluxAxis>\n");

		xml.append(linePrefix).append("</Point>\n");

		return xml.toString();
	}

}
