package sedplugin.sed.source;

import java.util.Iterator;
import java.util.NoSuchElementException;

import sedplugin.sed.SED;

/**
 * <p>
 * 	Lets iterating on all SEDs contained in a {@link SourcesGroup} object
 * 	whatever are its items (simple {@link SEDSource} or {@link SourcesGroup}).
 * </p>
 * 
 * <p><i><u>Note:</u> The remove() function is fully implemented !</i></p>
 * 
 * @author Gr&eacute;gory Mantelet (CDS)
 * @version 03/2011
 * 
 * @see SEDSource
 * @see SourcesGroup
 */
public final class SEDIterator implements Iterator<SED> {

	/** Sources on which this iterator must iterate. */
	protected Iterator<SEDSource> itSources;

	/** The source item which contains at least one SED. */
	protected SEDSource source = null;

	/** The iterator on the SEDs of the current source item.
	 * <b>If not <i>null</i>, this iterator has ALWAYS a next item.</b> */
	protected Iterator<SED> itSed = null;

	/** The source item whose the last SED has just been returned.
	 * It is kept for the {@link #remove()} function. */
	protected SEDSource previousSource = null;

	/** The iterator on SEDs of the previous source item.
	 * <b>If not <i>null</i>, this iterator has NEVER a next item.</b>
	 * It is kept for the {@link #remove()} function. */
	protected Iterator<SED> previousItSed = null;


	/**
	 * Builds a SEDIterator with the given {@link Iterator} of {@link SEDSource}.
	 * 
	 * @param sources	A {@link SEDSource} objects iterator.
	 * 
	 * @see #searchNextSource()
	 */
	public SEDIterator(Iterator<SEDSource> sources){
		itSources = sources;
		searchNextSource();
	}

	public boolean hasNext() {
		return (source != null);
	}

	/**
	 * @see java.util.Iterator#next()
	 * 
	 * @see #hasNext()
	 * @see SEDSource#getSED()
	 * @see #searchNextSource()
	 */
	public SED next() {
		if (!hasNext())
			throw new NoSuchElementException("No next SED !");

		SED nextSed = null;
		if (itSed == null)
			nextSed = source.getSED();
		else
			nextSed = itSed.next();

		if (itSed == null || !itSed.hasNext())
			searchNextSource();

		return nextSed;
	}

	/**
	 * @see java.util.Iterator#remove()
	 * 
	 * @see SEDSource#removeSED()
	 */
	public void remove() {
		if (previousSource == null)
			throw new IllegalStateException("No SED to remove ! You must call the function SEDIterator.next() to remove the returned SED from its SEDSource !");

		if (previousItSed == null)
			previousSource.removeSED();
		else
			previousItSed.remove();
	}

	/**
	 * <p>Searches the next {@link SEDSource} object which has at least one {@link SED}.</p>
	 * 
	 * <p>
	 * 	If a such {@link SEDSource} has been found, it is set into {@link #source}.
	 * 	Otherwise this attribute is set to <i>null</i>.
	 * </p>
	 * 
	 * <p>
	 * 	{@link #itSed} corresponds to an iterator on all {@link SED}s of the found {@link SEDSource}.
	 * 	It is <i>null</i> if the found {@link SEDSource} is not a {@link SourcesGroup}. In this case,
	 * 	that means the source contains only one {@link SED} which can be get by calling {@link SEDSource#getSED()}.
	 * </p>
	 */
	protected void searchNextSource(){
		previousSource = source;
		previousItSed = itSed;

		itSed = null;

		boolean hasSed = false;
		while(!hasSed && itSources.hasNext()){
			source = itSources.next();
			if (source instanceof SourcesGroup){
				itSed = ((SourcesGroup)source).getSEDs();
				hasSed = itSed.hasNext();
			}else
				hasSed = source.hasSED();
		}

		if (!hasSed){
			itSources = null;
			source = null;
			itSed = null;
		}
	}

}
