package sedplugin.sed.source;

import java.util.Iterator;
import java.util.Vector;

import sedplugin.sed.SED;

/**
 * Represents a group of sources (i.e. a drawing plan, a catalog or a file).
 * Thus a SourcesGroup can contain any instance of {@link SEDSource}, included a {@link SourcesGroup}.
 * 
 * @author Gr&eacute;gory Mantelet (CDS)
 * @version 04/2012
 * 
 * @see SEDSource
 */
public abstract class SourcesGroup implements SEDSource, Iterable<SEDSource>, SEDSourceListener {

	/** List of all sources gathered in this group. */
	protected final Vector<SEDSource> vSources = new Vector<SEDSource>(2, 10);

	/** Indicates whether the update ability is disabled/enabled in this group. */
	protected boolean frozen = false;

	/** Objects to notify when a SED is added into or removed from this SED source. */
	protected final Vector<SEDSourceListener> listeners = new Vector<SEDSourceListener>();

	public boolean addSEDSourceListener(SEDSourceListener l) {
		if (listeners.contains(l))
			return false;
		else
			return listeners.add(l);
	}

	public boolean removeSEDSourceListener(SEDSourceListener l) {
		return listeners.remove(l);
	}

	public final void sedAdded(SED newSed, SEDSource src) {
		for(SEDSourceListener l : listeners)
			l.sedAdded(newSed, src);
		src.freeze(frozen);
	}

	public final void sedRemoved(SED removedSed, SEDSource src) {
		for(SEDSourceListener l : listeners)
			l.sedRemoved(removedSed, src);
	}

	public void freeze(boolean freezeOrNot) {
		frozen = freezeOrNot;
		for(SEDSource src : vSources)
			src.freeze(freezeOrNot);
	}

	public boolean isFrozen() {
		return frozen;
	}

	/**
	 * <p>Updates the description of this source and then updates all contained sources ONLY IF THERE IS A MODIFICATION OF THE SOURCE.</p>
	 * 
	 * <p><i><u>Note:</u> Nothing is done is {@link #isFrozen()} returns </i>true<i>.</i></p>
	 * 
	 * @return <i>true</i> if all sources (included this one) has been successfully updated, <i>false</i> otherwise.
	 * 
	 * @see #update(boolean)
	 * @see SEDSource#update()
	 */
	public boolean update() {
		return update(false);
	}

	/**
	 * <p>Updates the description of this source and then updates all contained sources.</p>
	 * 
	 * <p><i><u>Note:</u> Nothing is done is {@link #isFrozen()} returns </i>true<i>.</i></p>
	 * 
	 * @param force		<i>true</i> to force the update even if there is no modification of the source, <i>false</i> otherwise.
	 * 
	 * @return <i>true</i> if all sources (included this one) has been successfully updated, <i>false</i> otherwise.
	 * 
	 * @see #primaryUpdate()
	 * @see SEDSource#update()
	 */
	public boolean update(boolean force) {
		// Update only this source:
		boolean updated = primaryUpdate(force);

		if (updated){
			// Update all contained and remaining sources:
			Iterator<SEDSource> it = vSources.iterator();
			while(it.hasNext()){
				SEDSource source = (SEDSource)it.next();
				if (!source.isFrozen())
					updated = source.update(force) && updated;
			}
		}

		return updated;
	}

	/**
	 * Updates only the description of this source. <b>No contained source must be updated !</b>
	 * 
	 * @param force		<i>true</i> to force the update even if there is no modification of the source, <i>false</i> otherwise.
	 * 
	 * @return	<i>true</i> if this source has been successfully updated, <i>false</i> otherwise.
	 * 
	 * @see #update()
	 */
	protected abstract boolean primaryUpdate(boolean force);

	/**
	 * Gets all sources gathered in this SourcesGroup.
	 * 
	 * @return	An iterator on all contained {@link SEDSource}.
	 */
	public final Iterator<SEDSource> getSources(){
		return vSources.iterator();
	}

	/**
	 * Gets all sources gathered in this SourcesGroup.
	 * 
	 * @return	An iterator on all contained {@link SEDSource}.
	 * 
	 * @see java.lang.Iterable#iterator()
	 * @see #getSources()
	 */
	public final Iterator<SEDSource> iterator(){
		return getSources();
	}

	/**
	 * Gets the specified source.
	 * 
	 * @param indSource	The index of the source to get.
	 * 
	 * @return			The corresponding {@link SEDSource} or <i>null</i> if the given index is incorrect.
	 */
	public final SEDSource getSource(int indSource){
		try{
			return vSources.get(indSource);
		}catch(ArrayIndexOutOfBoundsException ex){
			return null;
		}
	}

	/**
	 * Gets the number of all sources gathered in this SourcesGroup.
	 * 
	 * @return	The number of all sources.
	 */
	public final int getNbSources(){
		return vSources.size();
	}

	/**
	 * Gets the first SED (= getSED(0)).
	 * 
	 * @see #getSED(int)
	 */
	public SED getSED() {
		return getSED(0);
	}

	/**
	 * Gets - recursively - all {@link SED}s contained in this SourcesGroup.
	 * 
	 * @return	An iterator on all {@link SED}s.
	 */
	public Iterator<SED> getSEDs(){
		return new SEDIterator(vSources.iterator());
	}

	/**
	 * Gets - recursively - the specified {@link SED}.
	 * 
	 * @param indSED	The index of the {@link SED} to get.
	 * 
	 * @return			The corresponding {@link SED} or <i>null</i> if the given index is incorrect.
	 */
	public SED getSED(int indSED){
		Iterator<SED> it = getSEDs();
		SED sed = null;
		int i = -1;
		while(indSED > i++ && it.hasNext())
			sed = it.next();
		return sed;
	}

	public boolean hasSED() {
		return getSED(0) != null;
	}

	/**
	 * Gets - recursively - the number of all {@link SED}s.
	 * 
	 * @return	The number of all {@link SED}s.
	 */
	public int getNbSEDs(){
		Iterator<SED> it = getSEDs();
		int count = 0;
		while(it.hasNext()){
			it.next();
			count++;
		}
		return count;
	}

	/**
	 * Removes the first SED (= removeSED(0)).
	 * 
	 * @see #removeSED(int)
	 */
	public void removeSED() {
		removeSED(0);
	}

	/**
	 * Removes - recursively - the specified {@link SED}.
	 * 
	 * @param indSED	The index of the {@link SED} to remove.
	 */
	public void removeSED(int indSED){
		if (indSED < 0) return;

		Iterator<SED> it = getSEDs();
		int i = 0;

		while(it.hasNext()){
			it.next();
			if (indSED == i++){
				it.remove();
				return;
			}
		}
	}

	/**
	 * Removes all SED sources and their SEDs.
	 */
	public void clear(){
		for(SEDSource src : vSources){
			if (src instanceof SourcesGroup)
				((SourcesGroup)src).clear();
			else
				src.removeSED();
		}
		vSources.clear();
	}
}
