package sedplugin.sed.source.file;

import java.io.File;

import java.net.URL;

import java.util.Iterator;
import java.util.Vector;




import sedplugin.errors.ErrorsLog;
import sedplugin.errors.SEDException;
import sedplugin.sed.SED;
import sedplugin.sed.source.SEDSource;
import sedplugin.sed.source.SEDSourceListener;
import sedplugin.sed.unit.XAxis;
import sedplugin.sed.unit.YAxis;
import sedplugin.votable.SEDParser;


/**
 * A file or a URL which corresponds to a XML document representing one or several SEDs.
 * The XML document must respect the Photometry Data Model (cf Sebastien Derriere for more information about this model,
 * and see the class {@link SEDParser} which is supposed to be able to read this kind of file.
 * 
 * @author Gr&eacute;gory Mantelet (CDS)
 * @version 04/2012
 */
public class FileSource implements SEDSource {

	public final static String FILE_TYPE = "FILE";

	/** List of loaded SEDs. */
	protected final Vector<SED> seds = new Vector<SED>(1, 10);

	/** The file corresponding to the XML document which contains the SEDs to load. */
	protected final File file;

	protected long modificationDate;

	/** The URL toward the XML document which contains the SEDs to load. */
	protected final URL url;

	/* ************ */
	/* CONSTRUCTORS */
	/* ************ */
	/**
	 * Builds a FileSource with a file.
	 * 
	 * @param f	The file which contains the SEDs to load.
	 */
	public FileSource(File f) {
		file = f;
		url = null;
	}

	/**
	 * Builds a FileSource with a URL.
	 * 
	 * @param url	The URL toward the XML document which contains the SEDs to load.
	 */
	public FileSource(URL url) {
		this.url = url;
		file = null;
	}

	/* ***************** */
	/* GETTERS & SETTERS */
	/* ***************** */
	/**
	 * Gets the URL.
	 * 
	 * @return	Its URL.
	 */
	public URL getURL(){
		return url;
	}

	/**
	 * Gets the file.
	 * 
	 * @return	Its file.
	 */
	public File getFile(){
		return file;
	}

	/**
	 * Indicates whether this source is associated to a file or to a URL.
	 * 
	 * @return	<i>true</i> if this source is associated to a URL, <i>false</i> otherwise.
	 */
	public boolean isURL(){
		return url != null;
	}

	public String getName() {
		if (isURL())
			return url.getFile();
		else
			return file.getName();
	}

	public String getType() {
		return FILE_TYPE;
	}

	/* ***************** */
	/* INHERITED METHODS */
	/* ***************** */
	public boolean isModified(){
		return (!isURL() && modificationDate != file.lastModified());
	}

	public boolean addSEDSourceListener(SEDSourceListener l) {
		// TODO Auto-generated method stub
		return false;
	}

	public boolean removeSEDSourceListener(SEDSourceListener l) {
		// TODO Auto-generated method stub
		return false;
	}

	public boolean update(){
		return update(false);
	}

	public boolean update(final boolean force){
		if (!force && !isModified())
			return true;

		clear();

		SEDParser parser = new SEDParser(XAxis.getCurrentInstance(), YAxis.getCurrentInstance());
		try{
			// Parse the file/URL:
			if (isURL())
				parser.parse(url);
			else{
				parser.parse(file);
				modificationDate = file.lastModified();
			}

			// TODO Extracts also the information about the Spatial axis (center coordinates, radius, polygon points, ...) !

			// Add all the extracted SEDs to this source:
			Iterator<SED> it = parser.getSEDs();
			while(it.hasNext())
				seds.add(it.next());

			return true;
		}catch(Exception ex){
			ErrorsLog.getCurrentInstance().addException(new SEDException(ex, "Error at line "+parser.getLineNumber()+" and column "+parser.getColumnNumber()+" while parsing the file "+(isURL()?url:file)));
			/*System.err.println("((X)) ERROR AT LINE "+parser.getLineNumber()+" AND COLUMN "+parser.getColumnNumber()+": "+ex.getMessage());
			ex.printStackTrace();*/
		}
		return false;
	}

	public void highlight() { ; }

	public void freeze(boolean freezeOrNot) { ; }

	public boolean isFrozen() { return true; }

	/* ************** */
	/* SED MANAGEMENT */
	/* ************** */
	public SED getSED() {
		return seds.get(0);
	}

	/**
	 * Gets the specified SED.
	 * 
	 * @param indSED	The index of the SED to get.
	 * 
	 * @return			The corresponding SED.
	 */
	public SED getSED(int indSED){
		return seds.get(indSED);
	}

	/**
	 * Gets all loaded SEDs.
	 * 
	 * @return	An iterator on all loaded SEDs.
	 */
	public Iterator<SED> getSEDs(){
		return seds.iterator();
	}

	/**
	 * Gets the number of loaded SEDs.
	 * 
	 * @return	The number of loaded SEDs.
	 */
	public int getNbSEDs(){
		return seds.size();
	}

	public boolean hasSED() {
		return !seds.isEmpty();
	}

	public void removeSED() {
		seds.remove(0);
	}

	/**
	 * Removes the specified SED.
	 * 
	 * @param indSED	The index of the SED to remove.
	 */
	public void removeSED(int indSED){
		seds.remove(indSED);
	}

	/**
	 * Removes all loaded SEDs.
	 */
	public void clear(){
		seds.clear();
	}

	/* **************** */
	/* EXPORT FUNCTIONS */
	/* **************** */
	public String toASCIIFormat(String sep) { return null; }
	public String toVOTableFormat(String linePrefix) { /* TODO Gets spatial references of this source or other additional information. See Circle for an example. */ return null; }
	public String toXMLFormat(String linePrefix) { /* TODO Gets spatial references of this source or other additional information. See Circle for an example. */ return null; }

	/* ************** */
	/* TODO SELF TEST */
	/* ************** */
	public static void main(String[] args) throws Exception {
		YAxis.getCurrentInstance().setQuantity(YAxis.FLUX);

		System.out.println("##### FILE CASE #####");
		FileSource src = new FileSource(new File("photometry_examples/Examples Photometry/03-12-2010/UGC2826_IRAS_phot.xml"));
		System.out.println(src.getName()+" => "+src.getNbSEDs());
		src.update();
		System.out.println("\n"+src.getName()+" => "+src.getNbSEDs());

		System.out.println("\n##### URL CASE #####");
		src = new FileSource(new URL("http://cdsarc.u-strasbg.fr/viz-bin/votable/test?-source=2MASS-PSC&-c=lmc&-out.add=_RAJ,_DEJ&-oc.form=dec&-out.meta=DhuL&-out.max=999999&-c.rm=1.0&-out=_RAJ,_DEJ,_sed1"));
		System.out.println(src.getName()+" => "+src.getNbSEDs());
		src.update();
		System.out.println("\n"+src.getName()+" => "+src.getNbSEDs());
	}

}
