package sedplugin.sed.unit;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.Vector;

import cds.astro.Converter;

import sedplugin.errors.ErrorsLog;
import sedplugin.errors.SEDException;

import sedplugin.errors.unit.ConversionException;
import sedplugin.errors.unit.UnknownUnitException;

import sedplugin.sed.SED;
import sedplugin.sed.SEDPoint;
import sedplugin.sed.ToolBox;

/**
 * Representation of an axis.
 * 
 * @author Gr&eacute;gory Mantelet (CDS)
 * @version 03/2012
 * 
 * @see XAxis
 * @see YAxis
 */
public abstract class AxisUnit implements Iterable<PhysicalQuantity> {

	/** Name of this axis. It may be used as identifier. */
	protected final String name;

	/** List of all possible quantities for this axis. */
	protected final ArrayList<PhysicalQuantity> physicalQuantities;

	protected PhysicalQuantity customQuantity = null;

	/** Index of the current quantity. 2 special values: {@link #NONE} (if no quantity is selected) and {@link #physicalQuantities}.size() (if the custom quantity is selected). */
	protected int currentQuantity;

	/** The index which indicates there is no current quantity (generally used when there is no more quantity in this axis). */
	public final static int NONE = -1;

	/** List of SEDs to notify when the current unit of a quantity is changed. */
	protected final Vector<SED> seds;

	/* ************ */
	/* CONSTRUCTORS */
	/* ************ */
	/**
	 * Builds an axis.
	 * 
	 * @param name	Name of the axis.
	 */
	public AxisUnit(String name){
		this.name = name;
		physicalQuantities = new ArrayList<PhysicalQuantity>(2);
		currentQuantity = NONE;
		seds = new Vector<SED>(10, 10);
	}

	/* ***************** */
	/* GETTERS & SETTERS */
	/* ***************** */
	/**
	 * Gets the name of this axis.
	 * 
	 * @return	Its name.
	 */
	public final String getName(){
		return name;
	}

	/**
	 * Gets the index of the selected physical quantity.
	 * 
	 * @return	Index of the selected quantity.
	 */
	public int getIndQuantity() {
		return currentQuantity;
	}

	/**
	 * Gets the indQuantity-th quantity of this axis.
	 * 
	 * @param indQuantity	The index of the physical quantity to get.
	 * 
	 * @return	The corresponding physical quantity.
	 */
	public final PhysicalQuantity getQuantity(int indQuantity){
		if (isCustomQuantity(indQuantity))
			return customQuantity;
		else
			return physicalQuantities.get(indQuantity);
	}

	/**
	 * Gets all the quantities of this axis.
	 * 
	 * @return	An iterator on all its physical quantities.
	 */
	public final Iterator<PhysicalQuantity> getQuantities(){
		return physicalQuantities.iterator();
	}

	/**
	 * Gets all the quantities of this axis.
	 * 
	 * @return	An iterator on all its physical quantities.
	 * 
	 * @see #getQuantities()
	 */
	public final Iterator<PhysicalQuantity> iterator(){
		return getQuantities();
	}

	/**
	 * Gets the number of all quantities of this axis.
	 * 
	 * @return	The number of all its quantities.
	 */
	public final int getNbQuantities(){
		return physicalQuantities.size() + (hasCustomUnit() ? 1 : 0);
	}

	/**
	 * Tells whether the current quantity corresponds to the given one.
	 * 
	 * @param quantityType	A quantity type.
	 * 
	 * @return				<i>true</i> if the current quantity is the same as the given one, <i>false</i> otherwise.
	 */
	public final boolean isQuantity(final int quantityType){
		return currentQuantity == quantityType;
	}

	/**
	 * Gets the current quantity of this axis.
	 * 
	 * @return	Its current quantity or <i>null</i> if there is none.
	 */
	public final PhysicalQuantity getQuantity(){
		if (currentQuantity == NONE)
			return null;
		else if (isCustomUnit())
			return customQuantity;
		else
			return physicalQuantities.get(currentQuantity);
	}

	/**
	 * <p>Sets the current quantity considering the given quantity name.</p>
	 * 
	 * <p><i><u>Note:</u> All associated SEDs are notified.</i></p>
	 * 
	 * @param quantity	The name of the quantity to select as current quantity.
	 * 
	 * @return	<i>true</i> if the current quantity has been changed.
	 * 
	 * @see #notifySEDs()
	 */
	public final boolean setQuantity(String quantity){
		int indQuantity = indexOf(quantity);
		if (indQuantity != NONE){
			currentQuantity = indQuantity;
			notifySEDs();
			return true;
		}
		return false;
	}

	/**
	 * Tells whether this axis has a custom unit.
	 * 
	 * @return	<i>true</i> if this axis has a custom unit, <i>false</i> otherwise.
	 */
	public final boolean hasCustomUnit(){
		return (customQuantity != null && customQuantity.getNbUnits() == 1);
	}

	/**
	 * Tells whether the selected unit is the one of the custom physical quantity.
	 * 
	 * @return	<i>true</i> if the selected unit is custom, <i>false</i> otherwise.
	 * 
	 * @see #isCustomQuantity(int)
	 */
	public final boolean isCustomUnit(){
		return isCustomQuantity(currentQuantity);
	}

	/**
	 * Tells whether the given unit symbol is the custom unit of this axis.
	 * 
	 * @param unitSymbol	A unit symbol.
	 * 
	 * @return	<i>true</i> if the custom unit of this axis corresponds to the given one, <i>false</i> otherwise.
	 */
	public final boolean isCustomUnit(final String unitSymbol){
		if (hasCustomUnit())
			return customQuantity.exists(unitSymbol);
		else
			return false;
	}

	/**
	 * Tells whether the given index corresponds to the index of the custom quantity.
	 * 
	 * @param indQuantity	A quantity index.
	 * 
	 * @return	<i>true</i> if the given index is the one of the custom quantity, <i>false</i> otherwise.
	 */
	public final boolean isCustomQuantity(final int indQuantity){
		return (hasCustomUnit() && indQuantity == physicalQuantities.size());
	}

	/**
	 * Tells whether the given physical quantity corresponds to the custom quantity.
	 * 
	 * @param quantity	A physical quantity.
	 * 
	 * @return	<i>true</i> if the given quantity is the custom one, <i>false</i> otherwise.
	 */
	public final boolean isCustomQuantity(final PhysicalQuantity quantity){
		return (hasCustomUnit() && quantity != null && customQuantity.equals(quantity));
	}

	/**
	 * Gets the custom unit of this axis.
	 * 
	 * @return	The custom unit.
	 */
	public final UnitDef getCustomUnit(){
		if (customQuantity == null)
			return null;
		else
			return customQuantity.getCurrentUnit();
	}

	/**
	 * <p>Sets the custom unit as the current one.</p>
	 * 
	 * <p><i><u>Note:</u> All associated SEDs are notified.</i></p>
	 * 
	 * @return	<i>true</i> if the current unit has been changed.
	 * 
	 * @see #notifySEDs()
	 */
	public final boolean setCustomUnit(){
		if (customQuantity != null)
			currentQuantity = physicalQuantities.size();
		else
			currentQuantity--;
		notifySEDs();
		return true;
	}

	/**
	 * <p>Sets the custom unit to the given one and as the current one.</p>
	 * 
	 * <p><i><u>Note:</u> All associated SEDs are notified.</i></p>
	 * 
	 * @param custom	The unit to set as custom.
	 * 
	 * @return	<i>true</i> if the current unit has been changed.
	 * 
	 * @see #setCustomUnit()
	 */
	public final boolean setCustomUnit(UnitDef custom){
		if (custom == null)
			customQuantity = null;
		else{
			try{
				customQuantity = new PhysicalQuantity("custom", custom);
			}catch(SEDException se){
				ErrorsLog.getCurrentInstance().addException(se);
				customQuantity = null;
			}
		}
		setCustomUnit();
		return true;
	}

	/**
	 * <p>Sets the current quantity of this axis.</p>
	 * 
	 * <p><i><u>Note:</u> All associated SEDs are notified.</i></p>
	 * 
	 * @param indQuantity	The index of the quantity to select as current quantity of this axis.
	 * 
	 * @return	<i>true</i> if the current quantity has been changed, <i>false</i> otherwise (generally if the given index is incorrect).
	 * 
	 * @see #notifySEDs()
	 */
	public final boolean setQuantity(int indQuantity){
		if (isCustomQuantity(indQuantity) || (indQuantity >= 0 && indQuantity < physicalQuantities.size())){
			currentQuantity = indQuantity;
			notifySEDs();
			return true;
		}
		return false;
	}

	/**
	 * Gets the current unit of this axis.
	 * 
	 * @return	Its current unit or <i>null</i> if there is no quantity in this axis..
	 */
	public final UnitDef getUnit(){
		if (currentQuantity == NONE)
			return null;
		else
			return getQuantity().getCurrentUnit();
	}

	/**
	 * Gets an array gathering all units of the current physical quantity.
	 * 
	 * @return	All available units for the current physical quantity.
	 */
	public UnitDef[] getAllUnits() {
		PhysicalQuantity quantity = getQuantity();
		int nbUnits = quantity.getNbUnits();
		UnitDef[] allUnits = new UnitDef[nbUnits];

		for(int i=0; i<nbUnits; i++)
			allUnits[i] = quantity.getUnit(i);

		return allUnits;
	}

	/**
	 * Sets the current unit of this axis.
	 * 
	 * @param indUnit	The index of the unit to select as current unit in the current physical quantity.
	 * 
	 * @return	<i>true</i> if the current unit has been changed, <i>false</i> otherwise.
	 */
	public final boolean setUnit(int indUnit){
		if (currentQuantity != NONE)
			return getQuantity().setCurrentUnit(indUnit);
		return false;
	}

	/**
	 * <p>Sets the current quantity and the current unit in this quantity.</p>
	 * 
	 * <p><i><u>Note:</u> All the associated SEDs are automatically notified.</i></p>
	 * 
	 * @param indQuantity	The index of the quantity to select as current physical quantity.
	 * @param indUnit		The index of the unit to select as current unit in the specified quantity.
	 * 
	 * @return				<i>true</i> if the current quantity and/or unit have been changed, <i>false</i> otherwise.
	 * 
	 * @see #notifySEDs()
	 */
	public final boolean setUnit(int indQuantity, int indUnit){
		if (isCustomQuantity(indQuantity) || (indQuantity >= 0 && indQuantity < physicalQuantities.size())){
			currentQuantity = indQuantity;
			boolean unitUpdated = getQuantity().setCurrentUnit(indUnit);
			if (!unitUpdated)
				notifySEDs();
			return true;
		}
		return false;
	}

	/**
	 * <p>
	 * 	Sets the current unit considering the given unit symbol.
	 * 	The unit is selected whatever is its corresponding physical quantity.
	 * </p>
	 * 
	 * <p><i><u>Note:</u> All the associated SEDs are automatically notified.</i></p>
	 * 
	 * @param unitSymbol	The symbol of the unit to select as current unit.
	 * 
	 * @return				<i>true</i> if the current unit has been changed,
	 * 						<i>false</i> otherwise (generally if the given unit symbol does not correspond to a unit of any contained physical quantity).
	 * 
	 * @see #searchUnit(String)
	 * @see #setUnit(int, int)
	 */
	public final boolean setUnit(String unitSymbol){
		try {
			int[] indexes = searchUnit(unitSymbol);
			return setUnit(indexes[0], indexes[1]);
		} catch (UnknownUnitException e) {
			return false;
		}
	}

	/* **************************** */
	/* QUANTITY/UNIT SEARCH METHODS */
	/* **************************** */
	/**
	 * Gets the physical quantity whose the name is given in parameter.
	 * 
	 * @param quantityName	The name of the quantity to search.
	 * 
	 * @return	The corresponding physical quantity or <i>null</i> if no contained quantity has this name.
	 */
	public final PhysicalQuantity getQuantity(String quantityName){
		int indQuantity = indexOf(quantityName);
		if (indQuantity != NONE)
			return (PhysicalQuantity)physicalQuantities.get(indQuantity);
		else
			return null;
	}

	/**
	 * Gets the index of the given physical quantity.
	 * 
	 * @param quantity	The quantity to search.
	 * 
	 * @return	The index of the given quantity in this axis, or {@link #NONE} if no quantity can be found.
	 */
	public final int indexOf(PhysicalQuantity quantity){
		if (quantity == null)
			return NONE;

		if (isCustomQuantity(quantity))
			return physicalQuantities.size();
		else
			return physicalQuantities.indexOf(quantity);
	}

	/**
	 * Gets the index of the quantity whose the name is given in parameter.
	 * 
	 * @param quantity	The name of the quantity to search.
	 * 
	 * @return	The index of the corresponding physical quantity or {@link #NONE} if no quantity can be found.
	 */
	public final int indexOf(String quantity){
		if (quantity == null)
			return NONE;

		int indQuantity = NONE;
		for(int i=0; indQuantity == NONE && i<physicalQuantities.size(); i++){
			if (((PhysicalQuantity)physicalQuantities.get(i)).getName().equals(quantity))
				indQuantity = i;
		}
		return indQuantity;
	}

	/**
	 * Tells whether the given physical quantity exists in this axis.
	 * 
	 * @param quantity	The quantity to search.
	 * 
	 * @return	<i>true</i> if the given quantity exists, <i>false</i> otherwise.
	 * 
	 * @see #indexOf(PhysicalQuantity)
	 */
	public final boolean exists(PhysicalQuantity quantity){
		return indexOf(quantity) != NONE;
	}

	/**
	 * Tells whether there is a quantity with the given name in this axis.
	 * 
	 * @param quantity	The name of the quantity to search.
	 * 
	 * @return	<i>true</i> if a corresponding quantity can be found, <i>false</i> otherwise.
	 * 
	 * @see #indexOf(String)
	 */
	public final boolean exists(String quantity) {
		return indexOf(quantity) != NONE;
	}

	/**
	 * Tells whether a unit with the given symbol exists in any quantity of this axis.
	 * 
	 * @param unitSymbol	The symbol of the unit to search.
	 * 
	 * @return	<i>true</i> if a corresponding unit can be found, <i>false</i> otherwise.
	 */
	public final boolean unitExists(String unitSymbol){
		// Search in the custom quantity if any:
		if (hasCustomUnit() && customQuantity.exists(unitSymbol))
			return true;

		// Search in the other quantities:
		Iterator<PhysicalQuantity> itQuantities = getQuantities();
		while(itQuantities.hasNext()){
			PhysicalQuantity quantity = (PhysicalQuantity)itQuantities.next();
			if (quantity.exists(unitSymbol))
				return true;
		}

		return false;
	}

	/**
	 * Gets the index of the quantity and the index of unit in this quantity whose the symbol is given in parameter.
	 * 
	 * @param unitSymbol	The symbol of the unit to search.
	 * 
	 * @return	An array with two integers: [0] = index of the quantity which contains the corresponding unit, [1] = index of the corresponding unit.
	 * 
	 * @throws UnknownUnitException	If no corresponding unit can be found.
	 */
	protected final int[] searchUnit(String unitSymbol) throws UnknownUnitException {
		int indQuantity = NONE, indUnit = NONE;

		// Search in the custom quantity, if any:
		if (hasCustomUnit()){
			indUnit = customQuantity.indexOf(unitSymbol);
			if (indUnit != NONE)
				return new int[]{physicalQuantities.size(), indUnit};
		}

		// Search in all the other quantities:
		for(int i=0; indQuantity == NONE && i<physicalQuantities.size(); i++){
			indUnit = getQuantity(i).indexOf(unitSymbol);
			if (indUnit != NONE)
				indQuantity = i;
		}

		// If not found, throw an exception:
		if (indQuantity == NONE)
			throw new UnknownUnitException(unitSymbol, "Searching the unit \""+unitSymbol+"\" into the axis \""+getName()+"\"");
		else
			return new int[]{indQuantity, indUnit};
	}

	/* *************************** */
	/* QUANTITY MANAGEMENT METHODS */
	/* *************************** */
	/**
	 * Adds the given quantity to this axis.
	 * 
	 * @param newQuantity	The quantity to add.
	 * 
	 * @return	<i>true</i> if the given quantity has been added, <i>false</i> otherwise (generally if a quantity with the same name already exists).
	 */
	public final boolean add(PhysicalQuantity newQuantity){
		if (newQuantity == null || physicalQuantities.contains(newQuantity))
			return false;

		boolean currentIsCustom = isCustomUnit();
		boolean added = physicalQuantities.add(newQuantity);
		if (added){
			newQuantity.addAxis(this);
			if (currentQuantity == NONE)
				currentQuantity = 0;
			else if (currentIsCustom)
				setCustomUnit();
		}
		return added;
	}

	/**
	 * Removes the given quantity from this axis.
	 * 
	 * @param quantity	The quantity to remove.
	 * 
	 * @return	<i>true</i> if the given quantity has been removed, <i>false</i> otherwise.
	 */
	public final boolean remove(PhysicalQuantity quantity){
		// If it is the custom quantity:
		if (isCustomQuantity(quantity)){
			setCustomUnit(null);
			return true;
		}

		// Otherwise:
		boolean currentIsCustom = isCustomUnit();
		boolean removed = physicalQuantities.remove(quantity);
		if (removed){
			quantity.removeAxis(this);
			if (currentIsCustom)
				setCustomUnit();
			else if (physicalQuantities.isEmpty()){
				currentQuantity = NONE;
				notifySEDs();
			}else if (currentQuantity >= physicalQuantities.size())
				setQuantity(currentQuantity-1);
		}

		return removed;
	}

	/**
	 * Removes the quantity whose the name is given in parameter.
	 * 
	 * @param quantity	The name of the quantity to remove.
	 * 
	 * @return	<i>true</i> if the corresponding quantity has been removed, <i>false</i> otherwise.
	 * 
	 * @see #indexOf(String)
	 */
	public final boolean remove(String quantity){
		int indQuantity = indexOf(quantity);
		if (indQuantity != NONE){
			boolean currentIsCustom = isCustomUnit();
			PhysicalQuantity q = physicalQuantities.remove(indQuantity);
			if (q != null){
				q.removeAxis(this);
				if (currentIsCustom)
					setCustomUnit();
				else if (physicalQuantities.isEmpty()){
					currentQuantity = NONE;
					notifySEDs();
				}else if (currentQuantity >= physicalQuantities.size())
					setQuantity(currentQuantity-1);
				return true;
			}
		}
		return false;
	}

	/**
	 * Removes the indQuantity-th quantity from this axis.
	 * 
	 * @param indQuantity	The index of the quantity to remove.
	 * 
	 * @return	<i>true</i> if the corresponding quantity has been removed, <i>false</i> otherwise.
	 */
	public final boolean remove(int indQuantity){
		// If it is the custom quantity:
		if (isCustomQuantity(indQuantity)){
			setCustomUnit(null);
			return true;

		}// Otherwise:
		else if (indQuantity >= 0 && indQuantity < physicalQuantities.size()){
			boolean currentIsCustom = isCustomUnit();
			PhysicalQuantity q = (PhysicalQuantity)physicalQuantities.remove(indQuantity);
			if (q != null){
				q.removeAxis(this);
				if (currentIsCustom)
					setCustomUnit();
				else if (physicalQuantities.isEmpty()){
					currentQuantity = NONE;
					notifySEDs();
				}else if (currentQuantity >= physicalQuantities.size())
					setQuantity(currentQuantity-1);
				return true;
			}
		}
		return false;
	}

	/* ****************** */
	/* CONVERSION METHODS */
	/* ****************** */
	/**
	 * Computes the conversion of the corresponding axis value of the given SED point in the current unit.
	 * 
	 * @param point			The SED point to convert.
	 * 
	 * @return				The converted axis value.
	 * 
	 * @throws ConversionException	If one conversion is impossible.
	 * @throws UnknownUnitException	If one or both original units of the given SED point are unknown.
	 * 
	 * @see #extractValueToConvert(SEDPoint)
	 * @see #extractSourceUnit(SEDPoint)
	 * 
	 * @see ToolBox#getAppropriateConverter(String, String)
	 * 
	 * @see #searchUnit(String)
	 * 
	 * @see PhysicalQuantity#convertToRef(double, int)
	 * @see #convertQuantity(double, int, int, SEDPoint)
	 * @see PhysicalQuantity#convertFromRef(double, int)
	 */
	public double convert(SEDPoint point) throws ConversionException, UnknownUnitException {
		double value = extractValueToConvert(point);
		double result = value;
		String unit = extractSourceUnit(point);

		// Searches the source unit (note: an un-managed unit is allowed):
		boolean knownSrcUnit = true;
		int[] srcIndexes = new int[0];
		try{
			srcIndexes = searchUnit(unit);
		}catch(UnknownUnitException uue){
			knownSrcUnit = false;
		}

		// Conversion in 3 steps: 1) fromUnit -> fromUnitRef ; 2) fromUnitRef(fromQuantity) ->  toUnitRef(toQuantity) ; 3) toUnitRef -> toUnit
		if (knownSrcUnit && !isCustomQuantity(srcIndexes[0]) && !isCustomUnit()) {
			result = getQuantity(srcIndexes[0]).convertToRef(result, srcIndexes[1]);
			result = convertQuantity(result, srcIndexes[0], currentQuantity, point);
			result = getQuantity().convertFromRef(result, getQuantity().currentUnit);

		}// Otherwise: Use the Francois Ochsenbein's library (Astrocoo) to convert the value:
		else{
			Converter converter = ToolBox.getAppropriateConverter(unit, getUnit().symbol);
			if (converter == null)
				throw new ConversionException("Destination quantity incorrect: the unit \""+getUnit().symbol+"\" is not compatible with \""+unit+"\" !", "Quantity conversion in the axis \""+getName()+"\"");
			else
				result = converter.convert(result);
		}

		// Throw an exception if there is an error in the result value:
		if (Double.isNaN(result))
			throw new ConversionException(value, unit, getUnit().symbol, "converted value is NaN");
		else if (Double.isInfinite(result))
			throw new ConversionException(value, unit, getUnit().symbol, "converted value is Infinite");

		return result;
	}

	/**
	 * Computes the conversion of the corresponding axis value of the given SED point in the given unit.
	 * 
	 * @param point			The SED point to convert.
	 * @param toUnit		The destination unit.
	 * 
	 * @return				The converted axis value.
	 * 
	 * @throws ConversionException	If one conversion is impossible.
	 * @throws UnknownUnitException	If one or both original units of the given SED point are unknown.
	 * 
	 * @see #extractValueToConvert(SEDPoint)
	 * @see #extractSourceUnit(SEDPoint)
	 * 
	 * @see ToolBox#getAppropriateConverter(String, String)
	 * 
	 * @see #searchUnit(String)
	 * 
	 * @see PhysicalQuantity#convertToRef(double, int)
	 * @see #convertQuantity(double, int, int, SEDPoint)
	 * @see PhysicalQuantity#convertFromRef(double, int)
	 */
	public double convert(SEDPoint point, String toUnit) throws ConversionException, UnknownUnitException {
		double value = extractValueToConvert(point);
		double result = value;
		String fromUnit = extractSourceUnit(point);

		// Searches the source unit (note: an un-managed unit is allowed):
		boolean knownSrcUnit = true;
		int[] srcIndexes = new int[0];
		try{
			srcIndexes = searchUnit(fromUnit);
		}catch(UnknownUnitException uue){
			knownSrcUnit = false;
		}

		// Searches the destination unit (warning: any un-managed unit is forbidden !):
		int[] destIndexes = searchUnit(toUnit);

		// Conversion in 3 steps: 1) fromUnit -> fromUnitRef ; 2) fromUnitRef(fromQuantity) ->  toUnitRef(toQuantity) ; 3) toUnitRef -> toUnit
		if (knownSrcUnit && !isCustomQuantity(srcIndexes[0]) && !isCustomQuantity(destIndexes[0])) {
			result = getQuantity(srcIndexes[0]).convertToRef(result, srcIndexes[1]);
			result = convertQuantity(result, srcIndexes[0], destIndexes[0], point);
			result = getQuantity(destIndexes[0]).convertFromRef(result, destIndexes[1]);

		}// Otherwise: Use the Francois Ochsenbein's library (Astrocoo) to convert the value:
		else{
			Converter converter = ToolBox.getAppropriateConverter(fromUnit, toUnit);
			if (converter == null)
				throw new ConversionException("Destination quantity incorrect: the unit \""+toUnit+"\" is not compatible with \""+fromUnit+"\" !", "Quantity conversion in the axis \""+getName()+"\"");
			else
				result = converter.convert(result);
		}

		// Throw an exception if there is an error in the result value:
		if (Double.isNaN(result))
			throw new ConversionException(value, fromUnit, toUnit, "converted value is NaN");
		else if (Double.isInfinite(result))
			throw new ConversionException(value, fromUnit, toUnit, "converted value is Infinite");

		return result;
	}

	/**
	 * Extracts the original value of the corresponding axis from the given SED point.
	 * 
	 * @param point	The SED point from which the original value must be extracted.
	 * 
	 * @return	The original value of the corresponding axis.
	 */
	protected abstract double extractValueToConvert(SEDPoint point);

	/**
	 * Extracts the original unit of the corresponding axis from the given SED point.
	 * 
	 * @param point	The SED point from which the original unit must be extracted.
	 * 
	 * @return	The original unit of the corresponding axis.
	 */
	protected abstract String extractSourceUnit(SEDPoint point);

	/**
	 * Converts the given value from the reference value of the given source quantity to the reference value of the given destination quantity.
	 * 
	 * @param value				The value to convert.
	 * @param indFromQuantity	The index of the source physical quantity.
	 * @param indToQuantity		The index of the destination physical quantity.
	 * @param point				The SED point from which the given value has been extracted.
	 * 
	 * @return					The converted value.
	 * 
	 * @throws ConversionException		If the conversion is impossible.
	 * @throws UnknownUnitException		If one or both given units are unknown.
	 */
	protected abstract double convertQuantity(double value, int indFromQuantity, int indToQuantity, SEDPoint point) throws ConversionException, UnknownUnitException;

	/* ************************ */
	/* SED NOTIFICATION METHODS */
	/* ************************ */
	/**
	 * Adds the given SED to the list of SEDs to notify when the current unit of this axis changes.
	 * 
	 * @param newSED	A new SED to notify when the current unit changes.
	 */
	public final void addSED(SED newSED){
		seds.add(newSED);
	}

	/**
	 * Removes the given SED from the list of SEDs to notify when the current unit of this axis changes.
	 * 
	 * @param sedToRemove	The SED to remove.
	 */
	public final void removeSED(SED sedToRemove){
		seds.remove(sedToRemove);
	}

	/**
	 * Clears the list of SEDs to notify when the current unit of this axis changes.
	 */
	public final void removeAllSEDs(){
		seds.clear();
	}

	/**
	 * Updates the corresponding axis value of all SEDs.
	 */
	protected abstract void notifySEDs();

	/* ***************** */
	/* INHERITED METHODS */
	/* ***************** */
	/**
	 * Two axis are equals only if their name are strictly equals.
	 * 
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj){
		return (obj instanceof AxisUnit) && (((AxisUnit)obj).getName().equals(getName()));
	}

	@Override
	public String toString(){
		return getQuantity().toString();
	}

}
