package sedplugin.sed.unit;

import java.util.Iterator;



import sedplugin.errors.ErrorsLog;
import sedplugin.errors.SEDException;
import sedplugin.errors.unit.ConversionException;
import sedplugin.errors.unit.UnknownUnitException;
import sedplugin.sed.SED;
import sedplugin.sed.SEDPoint;
import sedplugin.sed.unit.PhysicalQuantity.UnitEntry;

/**
 * <p>Ordinate axis of a SED.</p>
 * 
 * <p>
 * 	There are, by default, only five physical quantities: Intensity (reference unit: Jy/sr), Nu.I(Nu) (reference unit: W/m2/sr),
 *  Flux (reference unit: Jy), Temperature(RJ) (reference unit: mK) and Temperature(CMB) (reference unit: mK).
 * </p>
 * 
 * @author Gr&eacute;gory Mantelet (CDS)
 * @version 06/2012
 */
public class YAxis extends AxisUnit {

	/** Index of the Intensity physical quantity. */
	public static final int INTENSITY = 0;

	/** Index of the Nu.Inu physical quantity. */
	public static final int NU_Inu = 1;

	/** Index of the Flux physical quantity. */
	public static final int FLUX = 2;

	/** Index of the Temperature (RJ) physical quantity. */
	public static final int TEMPERATURE_RJ = 3;

	/** Index of the Temperature (CMB) physical quantity. */
	public static final int TEMPERATURE_CMB = 4;

	/** Value to use to go from arc-secondes to steradian: value[arcsec]*EQUIVALENCE_ARCSEC_SR = value[sr] */
	protected static final double EQUIVALENCE_ARCSEC_SR = Math.pow(((Math.PI/180)/3600), 2);

	/** The associated {@link XAxis}. To make some conversions, the ordinate axis of a SED must know the abscissa axis (i.e. Intensity -> Flux). */
	protected final XAxis xAxis;

	/** The current instance of the {@link YAxis} class. */
	private static YAxis currentYAxis = null;

	/* *********** */
	/* CONSTRUCTOR */
	/* *********** */
	/**
	 * <p>Builds a YAxis with its associated {@link XAxis}.</p>
	 * 
	 * <p><b><u>Important: </u>You must ensure that this YAxis is always used with the given XAxis.</b></p>
	 * 
	 * <p>The physical quantities are also initialized:</p>
	 * <ul>
	 * 	<li>Intensity (name={@link #INTENSITY} and reference unit=Jy/sr):
	 * 		<ul>
	 * 			<li>Jy/sr</li>
	 * 			<li>MJy/sr <i>(default)</i></li>
	 * 			<li>W/m2/sr/Hz</li>
	 * 			<li>erg/cm2/s/sr/Hz</li>
	 * 		</ul>
	 * 	</li>
	 * 	<li>Nu.I(Nu) (name={@link #NU_Inu} and reference unit= W/m2/sr):
	 * 		<ul>
	 * 			<li>W/m2/sr <i>(default)</i></li>
	 * 			<li>erg/cm2/s/sr</li>
	 * 		</ul>
	 * 	</li>
	 * 	<li>Flux (name={@link #FLUX} and reference unit=Jy):
	 * 		<ul>
	 * 			<li>Jy <i>(default)</i></li>
	 * 			<li>W/m2/Hz</li>
	 * 			<li>erg/cm2/s/Hz</li>
	 * 		</ul>
	 * 	</li>
	 * 	<li>Temperature (RJ) (name={@link #TEMPERATURE_RJ} and reference unit=mK):
	 * 		<ul>
	 * 			<li>mK <i>(default)</i></li>
	 * 		</ul>
	 * 	</li>
	 * 	<li>Temperature (CMB) (name={@link #TEMPERATURE_CMB} and reference unit=mK):
	 * 		<ul>
	 * 			<li>mK <i>(default)</i></li>
	 * 		</ul>
	 * 	</li>
	 * </ul>
	 * 
	 * @param xAxis	Its associated {@link XAxis}.
	 */
	public YAxis(XAxis xAxis) throws SEDException {
		super("Ordinate");

		this.xAxis = xAxis;

		// Create the intensity units:
		PhysicalQuantity iQuantity = new PhysicalQuantity("Intensity", "I", "I\u03BD", new UnitDef("Jy/sr"));
		UnitEntry[] iUnits = new UnitEntry[]{
				new UnitEntry(new UnitDef("MJy/sr"), 1e6),
				new UnitEntry(new UnitDef("W/m2/sr/Hz", "W/m²/sr/Hz"), 1e26),
				new UnitEntry(new UnitDef("erg/cm2/s/sr/Hz", "erg/cm²/s/sr/Hz"), 1e23)
		};
		iQuantity.add(iUnits);
		iQuantity.setCurrentUnit(1);
		add(iQuantity);

		// Create the NuInu units:
		PhysicalQuantity nQuantity = new PhysicalQuantity("Nu.I(Nu)", "Freq.I", "\u03BD.I", new UnitDef("W/m2/sr", "W/m²/sr"));
		UnitEntry[] nUnits = new UnitEntry[]{
				new UnitEntry(new UnitDef("erg/cm2/s/sr", "erg/cm²/s/sr"), 1e-3)
		};
		nQuantity.add(nUnits);
		add(nQuantity);

		// Create the flux units:
		PhysicalQuantity fQuantity = new PhysicalQuantity("Flux", "F", new UnitDef("Jy"));
		UnitEntry[] fUnits = new UnitEntry[]{
				new UnitEntry(new UnitDef("W/m2/Hz", "W/m²/Hz"), 1e26),
				new UnitEntry(new UnitDef("erg/cm2/s/Hz", "erg/cm²/s/Hz"), 1e23)
		};
		fQuantity.add(fUnits);
		add(fQuantity);

		// Create the temperature (RJ) units:
		add(new PhysicalQuantity("Temperature (RJ)", new UnitDef("mK [RJ]")));

		// Create the temperature (CMB) units:
		add(new PhysicalQuantity("Temperature (CMB)", new UnitDef("mK [CMB]")));
	}

	/**
	 * Builds a YAxis with its associated {@link XAxis} and with the given unit as the current one.
	 * 
	 * @param xAxis			Its associated {@link XAxis}.
	 * @param unitName		Name of the unit to set as the current one.
	 */
	public YAxis(final XAxis xAxis, final String unitName) throws SEDException {
		this(xAxis);
		if (!setUnit(unitName))
			setCustomUnit(new UnitDef(unitName));
	}

	/* ****** */
	/* GETTER */
	/* ****** */
	/**
	 * Gets the current instance of {@link YAxis}.
	 * 
	 * @return	The current instance of {@link YAxis}.
	 */
	public static final synchronized YAxis getCurrentInstance(){
		if (currentYAxis == null){
			try{
				currentYAxis = new YAxis(XAxis.getCurrentInstance());
			}catch(SEDException se){
				ErrorsLog.getCurrentInstance().addException(se);
			}
		}
		return currentYAxis;
	}

	/* ***************** */
	/* INHERITED METHODS */
	/* ***************** */
	@Override
	protected void notifySEDs() {
		Iterator<SED> it = seds.iterator();
		while(it.hasNext())
			((SED)it.next()).updateY();
	}

	@Override
	protected double extractValueToConvert(SEDPoint point){
		return point.getOriginalY();
	}

	@Override
	protected String extractSourceUnit(SEDPoint point){
		return point.getOriginalYUnit();
	}

	@Override
	protected double convertQuantity(double value, int indFromQuantity, int indToQuantity, SEDPoint point) throws ConversionException, UnknownUnitException {
		if (indFromQuantity == indToQuantity)
			return value;

		// TODO DEBUG Trace !
		if (SED.trace) System.out.println("** Convert quantity: "+getQuantity(indFromQuantity).getUnicodeSymbol()+" -> "+getQuantity(indToQuantity).getUnicodeSymbol());

		double intensity, cst;
		switch (indFromQuantity){
		case INTENSITY:
			switch(indToQuantity){
			case NU_Inu:
				return value * 1e-26 * getXInFrequency(point);
			case FLUX:
				return value * (point.getSurface()*EQUIVALENCE_ARCSEC_SR);
			case TEMPERATURE_RJ:
				cst = getCstRJ(getXInWavelength(point));
				if (cst == 0)
					throw new ConversionException("RJ constant is null", "While converting "+value+getQuantity(indFromQuantity).getSymbol()+" into "+getQuantity(indToQuantity).getSymbol());
				return (value*1e-6)/cst;

			case TEMPERATURE_CMB:
				cst = getCstCMB(getXInWavelength(point));
				if (cst == 0)
					throw new ConversionException("CMB constant is null", "While converting "+value+getQuantity(indFromQuantity).getSymbol()+" into "+getQuantity(indToQuantity).getSymbol());
				return (value*1e-6)/cst;

			}
			break;
		case NU_Inu:
			intensity = getXInFrequency(point) * 1e-26;
			intensity = (intensity == 0)?Double.NaN:(value/intensity);

			switch(indToQuantity){
			case INTENSITY:
				return intensity;
			case FLUX:
				return convertQuantity(intensity, INTENSITY, FLUX, point);
			case TEMPERATURE_RJ:
				return convertQuantity(intensity, INTENSITY, TEMPERATURE_RJ, point);
			case TEMPERATURE_CMB:
				return convertQuantity(intensity, INTENSITY, TEMPERATURE_CMB, point);
			}
			break;
		case FLUX:
			intensity = (point.getSurface() == 0)?Double.NaN:(value/(point.getSurface()*EQUIVALENCE_ARCSEC_SR));
			switch(indToQuantity){
			case INTENSITY:
				return intensity;
			case NU_Inu:
				return convertQuantity(intensity, INTENSITY, NU_Inu, point);
			case TEMPERATURE_RJ:
				return convertQuantity(intensity, INTENSITY, TEMPERATURE_RJ, point);
			case TEMPERATURE_CMB:
				return convertQuantity(intensity, INTENSITY, TEMPERATURE_CMB, point);
			}
			break;
		case TEMPERATURE_RJ:
			intensity = value * getCstRJ(getXInWavelength(point)) * 1e6;
			switch(indToQuantity){
			case INTENSITY:
				return intensity;
			case NU_Inu:
				return convertQuantity(intensity, INTENSITY, NU_Inu, point);
			case FLUX:
				return convertQuantity(intensity, INTENSITY, FLUX, point);
			case TEMPERATURE_CMB:
				return convertQuantity(intensity, INTENSITY, TEMPERATURE_CMB, point);
			}
			break;
		case TEMPERATURE_CMB:
			intensity = value * getCstCMB(getXInWavelength(point)) * 1e6;
			switch(indToQuantity){
			case INTENSITY:
				return intensity;
			case NU_Inu:
				return convertQuantity(intensity, INTENSITY, NU_Inu, point);
			case FLUX:
				return convertQuantity(intensity, INTENSITY, FLUX, point);
			case TEMPERATURE_RJ:
				return convertQuantity(intensity, INTENSITY, TEMPERATURE_RJ, point);
			}
			break;
		default:
			throw new ConversionException("Source quantity incorrect: there is no "+indFromQuantity+"th physical quantity !", "Quantity conversion in the axis \""+getName()+"\"");
		}
		throw new ConversionException("Destination quantity incorrect: there is no "+indToQuantity+"th physical quantity !", "Quantity conversion in the axis \""+getName()+"\"");
	}

	/**
	 * Extracts the X value from the given point and converts it in Hz.
	 * 
	 * @param point			The SED point from which the X value must be extracted.
	 * 
	 * @return				The corresponding X value (in Hz).
	 *
	 * @throws ConversionException	If one conversion is impossible.
	 * @throws UnknownUnitException	If one or both original units of the given SED point are unknown.
	 * 
	 * @see XAxis#convert(double, String, String)
	 */
	protected final double getXInFrequency(SEDPoint point) throws ConversionException, UnknownUnitException {
		boolean oldTrace = SED.trace;
		SED.trace = false;
		double result = xAxis.convert(point.getOriginalX(), point.getOriginalXUnit(), "Hz");
		SED.trace = oldTrace;
		return result;
	}

	/**
	 * Extracts the X value from the given point and converts it in um.
	 * 
	 * @param point			The SED point from which the X value must be extracted.
	 * 
	 * @return				The corresponding X value (in um).
	 * 
	 * @throws ConversionException	If one conversion is impossible.
	 * @throws UnknownUnitException	If one or both original units of the given SED point are unknown.
	 * 
	 * @see XAxis#convert(double, String, String)
	 */
	protected final double getXInWavelength(SEDPoint point) throws ConversionException, UnknownUnitException {
		boolean oldTrace = SED.trace;
		SED.trace = false;
		double result = xAxis.convert(point.getOriginalX(), point.getOriginalXUnit(), "um");
		SED.trace = oldTrace;
		return result;
	}

	/**
	 * <p>Computes the RJ (Rayleigh-Jeans) constant (part of the conversion which depends only from the wavelength (um)):</p>
	 * <p>I[MJy/sr] = T[mK] * RJ_CST(lambda[um])</p>
	 * 
	 * @param lambda	Wavelength (in um).
	 * 
	 * @return			The RJ constant.
	 * 
	 * @throws ConversionException	If the computed constant is NaN or Infinite.
	 */
	protected final double getCstRJ(double lambda) throws ConversionException {
		double result = ((1e-3)*2*(1.38e-23)*(1e20))/Math.pow((lambda*(1e-6)), 2);

		if (Double.isNaN(result) || Double.isInfinite(result))
			throw new ConversionException("Problem when computing the RJ constant in function of lambda ("+lambda+"um): the result is "+result+".", "Conversion from/to the temperature according to the Rayleigh-Jeans law");

		return result;
	}

	/**
	 * <p>Computes the CMB (Cosmic Microwave Background) constant (part of the conversion which depends only from the wavelength (um)):</p>
	 * <p>I[MJy/sr] = T[mK] * CMB_CST(lambda[um]) = T[mK] * 1e-3 * dB/dT</p>
	 * 
	 * @param lambda	Wavelength (in um).
	 * 
	 * @return			The CMB constant.
	 * 
	 * @throws ConversionException	If the computed constant is NaN or Infinite.
	 */
	protected final double getCstCMB(double lambda) throws ConversionException {
		// FIRST METHOD: theoretically right, but the result is not correct !
		//		double h = 6.6260e-34;			// m².kg/s
		//		double c = 299792458;			// m/s
		//		double k = 1.3806505e-23;		// m².kg/s²/K
		//		double T = 2.725;				// K

		//		double var = (h*c)/(lambda*k*T);

		//		double result = ((2*Math.pow(h,2)*Math.pow(c,7)) / (k*Math.pow(lambda, 4)*Math.pow(T, 2)*(Math.exp(var)-1)*(1-Math.exp(-var)))) * ((1e-3)/(2.41798e22));

		// SECOND METHOD: all factors not-dependant from lambda have been already computed:
		double result = ((7.651e13)/Math.pow(lambda,4.0)) * (1.0/((Math.exp(5280.0/lambda)-1.0) * (1-Math.exp(-5280.0/lambda))));

		if (Double.isNaN(result) || Double.isInfinite(result))
			throw new ConversionException("Problem when computing the CMB constant in function of lambda ("+lambda+"um): the result is "+result+".", "Conversion from/to the temperature according to the Cosmic Microwave Background radiation");

		return result;
	}

	/* ************** */
	/* TODO SELF TEST */
	/* ************** */
	public static void main(String[] args) throws Exception {
		XAxis xAxis = new XAxis();
		YAxis yAxis = new YAxis(xAxis);
		System.out.println("AXIS["+yAxis.getName()+"]: "+yAxis);
		Iterator<PhysicalQuantity> itQuantities = yAxis.getQuantities();
		while(itQuantities.hasNext())
			((PhysicalQuantity)itQuantities.next()).display();
	}

}
