package sedplugin.votable;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.net.URL;
import java.util.HashMap;

import org.kxml2.io.KXmlParser;

import cds.savot.common.Markups;

/**
 * General algorithm to read a VOTable file using {@link KXmlParser}.
 * 
 * @author Gr&eacute;gory Mantelet (CDS)
 * @version 06/2012
 */
public abstract class VOTableParser implements Markups {
	protected String currentTag;
	protected String tagStack;

	private KXmlParser parser;
	private InputStream input;

	protected VOTableParser(){
		input = null;
	}

	protected VOTableParser(InputStream voTableInput) throws Exception {
		parse(voTableInput);
	}

	protected VOTableParser(URL voTableUrl) throws Exception {
		parse(voTableUrl);
	}

	protected VOTableParser(File voTableFile) throws Exception {
		parse(voTableFile);
	}

	public void parse() throws Exception {
		parse(input);
	}

	public void parse(URL voTableUrl) throws Exception {
		parse(new BufferedInputStream(voTableUrl.openStream()));
	}

	public void parse(File voTableFile) throws Exception {
		if (voTableFile.exists() && voTableFile.canRead())
			parse(new BufferedInputStream(new FileInputStream(voTableFile)));
		else
			throw new Exception("Impossible to open/read the file \""+voTableFile+"\" !");
	}

	public void parse(InputStream voTableInput) throws Exception {
		input = voTableInput;

		// Quick check of the input:
		if (input == null)
			throw new Exception("Impossible to parse a VOTable without a valid input (different from NULL) !");

		// Reset the parser:
		reset();

		// Build the parser:
		parser = new KXmlParser();
		//		parser.setFeature(KXmlParser.FEATURE_PROCESS_NAMESPACES, true);
		parser.setInput(input, "UTF-8");

		// Parse the given file:
		int eventType = parser.getEventType();
		while (eventType != KXmlParser.END_DOCUMENT) {
			// START DOCUMENT:
			if(eventType == KXmlParser.START_DOCUMENT)
				startDocument();

			// END DOCUMENT:
			else if(eventType == KXmlParser.END_DOCUMENT)
				endDocument();

			// START TAG:
			else if(eventType == KXmlParser.START_TAG) {
				// Get node name and update the tag stack:
				currentTag = parser.getName().trim();
				tagStack += ((tagStack.length() == 0)?"":" ")+currentTag;
				// Get attributes (if any):
				HashMap<String, String> attributes = new HashMap<String, String>(parser.getAttributeCount());
				for(int i=0; i<parser.getAttributeCount(); i++)
					attributes.put(parser.getAttributeName(i).trim(), parser.getAttributeValue(i).trim());
				// Call the appropriate method:
				if (currentTag.equals(VOTABLE))
					startVoTable(attributes);
				else if (currentTag.equals(RESOURCE))
					startResource(attributes);
				else if (currentTag.equals(TABLE))
					startTable(attributes);
				else if (currentTag.equals(GROUP))
					startGroup(attributes);
				else if (currentTag.equals(PARAM))
					startParam(attributes);
				else if (currentTag.equals(FIELD))
					startField(attributes);
				else if (currentTag.equals(FIELDREF))
					startFieldRef(attributes);
				else if (currentTag.equals(TABLEDATA))
					startTableData();
				else if (currentTag.equals(TR))
					startTR(attributes);
				else if (currentTag.equals(TD))
					startTD(attributes);
				else if (currentTag.equals(DESCRIPTION))
					startDescription(attributes);

				// END TAG:
			}else if(eventType == KXmlParser.END_TAG) {
				// Call the appropriate method:
				if (currentTag.equals(VOTABLE))
					endVoTable();
				else if (currentTag.equals(RESOURCE))
					endResource();
				else if (currentTag.equals(TABLE))
					endTable();
				else if (currentTag.equals(GROUP))
					endGroup();
				else if (currentTag.equals(PARAM))
					endParam();
				else if (currentTag.equals(FIELD))
					endField();
				else if (currentTag.equals(FIELDREF))
					endFieldRef();
				else if (currentTag.equals(TABLEDATA))
					endTableData();
				else if (currentTag.equals(TR))
					endTR();
				else if (currentTag.equals(TD))
					endTD();
				else if (currentTag.equals(DESCRIPTION))
					endDescription();
				// Update the tag stack and the current tag name:
				if (tagStack.length() > 0){
					int indLastSpace = tagStack.lastIndexOf(" ");
					if (indLastSpace < 0){
						currentTag = tagStack;
						tagStack = "";
					}else{
						tagStack = tagStack.substring(0, indLastSpace);
						currentTag = tagStack.substring(tagStack.lastIndexOf(" ")+1);
					}
				}

				// END TEXT:
			}else if(eventType == KXmlParser.TEXT) {
				// Get the text content:
				String text = parser.getText().trim();
				// Call the appropriate method:
				if (text.length() > 0){
					if (currentTag.equals(TD))
						textTD(text);
					else if (currentTag.equals(DESCRIPTION))
						textDescription(text);
				}
			}

			// GO TO NEXT TAG:
			eventType = parser.next();
		}

		endDocument();
	}

	protected void reset(){
		parser = null;
		currentTag = "";
		tagStack = "";
	}

	public int getLineNumber(){
		return parser.getLineNumber();
	}

	public int getColumnNumber(){
		return parser.getColumnNumber();
	}

	protected abstract void startDocument() throws Exception;
	protected abstract void endDocument() throws Exception;

	protected abstract void startVoTable(HashMap<String, String> attributes) throws Exception;
	protected abstract void endVoTable() throws Exception;

	protected abstract void startResource(HashMap<String, String> attributes) throws Exception;
	protected abstract void endResource() throws Exception;

	protected abstract void startTable(HashMap<String, String> attributes) throws Exception;
	protected abstract void endTable() throws Exception;

	protected abstract void startGroup(HashMap<String, String> attributes) throws Exception;
	protected abstract void endGroup() throws Exception;

	protected abstract void startParam(HashMap<String, String> attributes) throws Exception;
	protected abstract void endParam() throws Exception;

	protected abstract void startField(HashMap<String, String> attributes) throws Exception;
	protected abstract void endField() throws Exception;

	protected abstract void startFieldRef(HashMap<String, String> attributes) throws Exception;
	protected abstract void endFieldRef() throws Exception;

	protected abstract void startTableData() throws Exception;
	protected abstract void endTableData() throws Exception;

	protected abstract void startTR(HashMap<String, String> attributes) throws Exception;
	protected abstract void endTR() throws Exception;

	protected abstract void startTD(HashMap<String, String> attributes) throws Exception;
	protected abstract void textTD(String text) throws Exception;
	protected abstract void endTD() throws Exception;

	protected abstract void startDescription(HashMap<String, String> attributes) throws Exception;
	protected abstract void textDescription(String text) throws Exception;
	protected abstract void endDescription() throws Exception;

}
